from collections import defaultdict
from pathlib import Path

import pytest

from pyaerocom import utils


def test_print_file(tmp_path: Path, capsys):
    path = tmp_path / "file.txt"
    text = "Blaaaa\nBlub\n"
    path.write_text(text)

    utils.print_file(path)
    captured = capsys.readouterr()
    assert captured.out == text


def test_print_file_error(tmp_path: Path):
    path = tmp_path / "file.txt"
    assert not path.exists()
    with pytest.raises(IOError) as e:
        utils.print_file(path)
    assert str(e.value) == "File not found..."

    folder = tmp_path / "not_a_file"
    folder.mkdir()
    with pytest.raises(ValueError) as e:
        utils.print_file(folder)
    assert str(e.value) == f"{folder} is not a file"


@pytest.mark.parametrize(
    "kwargs,tabshape",
    [
        pytest.param(
            dict(model_ids="TM5-met2010_CTRL-TEST", vars_or_var_patterns="abs550*"),
            (2, 11),
            id="varinfo abs550*",
        ),
        pytest.param(
            dict(model_ids="TM5-met2010_CTRL-TEST", vars_or_var_patterns="*550*"),
            (4, 11),
            id="varinfo *550*",
        ),
        pytest.param(
            dict(
                model_ids="TM5-met2010_CTRL-TEST", vars_or_var_patterns="od550aer", read_data=True
            ),
            (2, 11),
            id="read od550aer",
        ),
    ],
)
def test_create_varinfo_table(kwargs, tabshape):
    df = utils.create_varinfo_table(**kwargs)
    assert df.shape == tabshape


@pytest.mark.parametrize(
    "kwargs,error",
    [
        pytest.param(
            dict(),
            "missing 2 required positional arguments: 'model_ids' and 'vars_or_var_patterns'",
            id="no args",
        ),
        pytest.param(
            dict(model_ids="TM5-met2010_CTRL-TEST"),
            "missing 1 required positional argument: 'vars_or_var_patterns'",
            id="no vars_or_var_patterns",
        ),
        pytest.param(
            dict(vars_or_var_patterns="od550aer"),
            "missing 1 required positional argument: 'model_ids'",
            id="no model_ids",
        ),
    ],
)
def test_create_varinfo_table_error(kwargs, error):
    with pytest.raises(TypeError) as e:
        utils.create_varinfo_table(**kwargs)
    assert str(e.value).endswith(error)


def test_recursive_default_dict_1():
    d = utils.recursive_defaultdict()

    d["A"]["B"]["C"]["D"]["E"] = "Hello world"

    assert d["A"]["B"]["C"]["D"]["E"] == "Hello world"


def test_recursive_default_dict_2():
    input_dict = {
        "A": {"AA": {"AAA": 5}, "AB": 4},
        "B": {"BA": 1, "BB": {"BBA": 4, "BBB": 5}},
    }

    d = utils.recursive_defaultdict(input_dict)

    def check_dict_keys_and_values(d1, d2):
        if not isinstance(d1, dict | defaultdict):
            assert d1 == d2
            return

        assert set(d1.keys()) == set(d2.keys())

        for k in d1.keys():
            check_dict_keys_and_values(d1[k], d2[k])

    check_dict_keys_and_values(input_dict, d)



import logging

import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom.colocation.colocated_data import ColocatedData

logger = logging.getLogger(__name__)


def min_periods_max(x: np.ndarray, /, min_periods=1) -> float:
    """Calculates the max of a 1-dimensional array, returning
    nan if not a minimum count of valid values exist.

    :param x: 1-dimensional ndarray.
    :param min_periods: minimum required non-nan values, defaults to 1
    :return: A single value, which is either nan or a float.
    """
    if x.ndim != 1:
        raise ValueError(f"Unexpected number of dimensions. Got {x.ndim}, expected 1.")

    length = np.sum(~np.isnan(x))
    if length < min_periods:
        return np.nan

    return np.nanmax(x)


def mda8_colocated_data(coldat: ColocatedData, /, obs_var: str, mod_var: str) -> ColocatedData:
    """Applies the mda8 calculation to a colocated data object,
    returning the new colocated data object.

    :param data: The colocated data object.
    :return: Colocated data object containing
    """
    if not isinstance(coldat, ColocatedData):
        raise ValueError(f"Unexpected type {type(coldat)}. Expected ColocatedData")

    if coldat.ts_type != "hourly":
        raise ValueError(f"Expected hourly timeseries. Got {coldat.ts_type}.")

    # TODO: Currently order of dims matter in the implementation, so this check is
    # stricter than it probably should be.
    if coldat.dims != ("data_source", "time", "station_name"):
        raise ValueError(
            f"Unexpected dimensions. Got {coldat.dims}, expected ['data_source', 'time', 'station_name']."
        )

    cd = ColocatedData(_calc_mda8(coldat.data))
    cd.data.attrs["var_name"] = [obs_var, mod_var]
    cd.metadata["var_name_input"] = [obs_var, mod_var]

    return cd


def _calc_mda8(data: xr.DataArray) -> xr.DataArray:
    """Calculates the daily max 8h average for an array:

    :param data: The DataArray for which to calculate the mda8. Input
    should be a DataArray with dimensions ["data_source", "time", "station_name"]
    (ie. the format of ColocatedData.data) representing hourly data.
    :return: Equivalently structured DataArray, resampled along the "time"
    dimension.

    Note:
    -----
    The calculation for mda8 is defined as follows:
    > Eight hours values:         75 % of values (i.e. 6 hours)

    > Maximum daily 8-hour mean:  75 % of the hourly running eight hour
                             averages (i.e. 18 eight hour averages per
                             day)

    > The maximum daily eight hour mean concentration will be selected by examining
    > eight hour running averages, calculated from hourly data and updated each hour.
    > Each eight hour average so calculated will be assigned to the day on which it
    > ends i.e. the first calculation period for any one day will be the period from
    > 17:00 on the previous day to 01:00 on that day; the last calculation period for
    > any one day will be the period from 16:00 to 24:00 on that day.

    Note:
    -----
    Calculated values will only be returned for days which have at least one datapoint
    in the input dataarray to ensure that the ts does not expand.

    See also:
    ---------
    https://eur-lex.europa.eu/legal-content/EN/TXT/HTML/?uri=CELEX:32008L0050#ntc3-L_2008152EN.01003001-E0003 (Annex XI)
    """
    mda8 = _daily_max(_rolling_average_8hr(data))

    mda8.attrs["ts_type"] = "daily"

    # Ensure time dimension represents the midpoint of the interval.
    mda8.coords.update({"time": mda8.get_index("time") + pd.tseries.frequencies.to_offset("12h")})

    # Keep only values for days that existed in the original time series.
    mda8 = mda8.sel(
        time=np.isin(mda8.coords["time.date"].values, np.unique(data.coords["time.date"].values))
    )

    return mda8


def _rolling_average_8hr(arr: xr.DataArray) -> xr.DataArray:
    # Xarray labels the data right based on the last data point in the period for rolling.
    return arr.rolling(time=8, center=False, min_periods=6).mean()


def _daily_max(arr: xr.DataArray) -> xr.DataArray:
    return arr.resample(time="24H", origin="start_day", label="left", offset="1h").reduce(
        lambda x, axis: np.apply_along_axis(min_periods_max, 1, x, min_periods=18)
    )



from pytest import approx

from pyaerocom.colocation.colocation_utils import colocate_gridded_ungridded
from pyaerocom.griddeddata import GriddedData
from pyaerocom.io.mscw_ctm.reader import ReadMscwCtm
from pyaerocom.ungriddeddata import UngriddedData
from tests.fixtures.data_access import TEST_DATA
from tests.fixtures.stations import create_fake_station_data

ROOT = TEST_DATA["MODELS"].path


def test_read_emep_colocate_projection():
    reader = ReadMscwCtm(data_dir=str(ROOT / "emep4no20240630"))
    data_emep = reader.read_var("concpm10", ts_type="hourly")
    assert isinstance(data_emep, GriddedData)

    S1 = create_fake_station_data(
        "concpm10",
        {"concpm10": {"units": "ug m-3"}},
        1,
        "2024-06-30T23:30:00",
        "2024-07-01T05:30:00",
        "h",
        {"ts_type": "hourly"},
    )
    S1.longitude = 10
    S1.latitude = 60
    S1.station_name = "S1"
    S2 = create_fake_station_data(
        "concpm10",
        {"concpm10": {"units": "ug m-3"}},
        2,
        "2024-06-30T23:30:00",
        "2024-07-01T05:30:00",
        "h",
        {"ts_type": "hourly"},
    )
    S2.longitude = 10.5
    S2.latitude = 60.5
    S2.station_name = "S2"
    ug = UngriddedData.from_station_data([S1, S2])
    cd = colocate_gridded_ungridded(data_emep, ug)
    assert cd.data.dims == ("data_source", "time", "station_name")
    assert cd.data[0, 0, 0] == 1
    assert cd.data[0, 0, 1] == 2
    assert cd.data[1, 0, 0] == approx(2.558, abs=1e-3)
    assert cd.data[1, 0, 1] == approx(2.444, abs=1e-3)



import pytest

from pyaerocom.molmasses import get_mmr_to_vmr_fac, get_molmass, get_species


@pytest.mark.parametrize(
    "var_name,species",
    [
        ("air_dry", "air_dry"),
        ("concno2", "no2"),
        ("sconcso2", "so2"),
        ("vmro3", "o3"),
        ("mmro3", "o3"),
        ("wetso4", "so4"),
        ("concNnh4", "nh4"),
        ("concNnh3", "nh3"),
        ("concNtno3", "no3"),
        ("proxydryno2", "no2"),
        ("proxywetno2", "no2"),
    ],
)
def test_get_species(var_name: str, species: str):
    assert get_species(var_name) == species


@pytest.mark.parametrize(
    "var_name,molmass",
    [
        ("air_dry", 28.9647),
        ("concno2", 46.0055),
        ("sconcso2", 64.066),
        ("vmro3", 48),
        ("mmro3", 48),
        ("wetso4", 96.06),
        ("concNnh4", 18.039),
        ("concNnh3", 17.031),
        ("concNtno3", 62.0045),
        ("proxydryno2", 46.0055),
    ],
)
def test_get_molmass(var_name: str, molmass: float):
    assert get_molmass(var_name) == molmass


@pytest.mark.parametrize(
    "var_name,result",
    [
        ("mmro3", 0.60343125),
        ("conco3", 0.60343125),
    ],
)
def test_get_mmr_to_vmr_fac(var_name: str, result: float):
    assert get_mmr_to_vmr_fac(var_name) == pytest.approx(result, rel=1e-3)



#!/usr/bin/env python3

"""simple script to generate a small enough test data set for the EBAS obs network
Works only if the user has access to the standard EBAS data path at Met Norway
"""

import os
import shutil
from pathlib import Path

import simplejson

import pyaerocom as pya

# import pyaerocom.access_testdata as td
from pyaerocom.access_testdata import AccessTestData

# from getpass import getuser
#
# if getuser() == 'jonasg':
#     ebas_local = os.path.join(pya.const.OUTPUTDIR, 'data/obsdata/EBASMultiColumn/data')
#     assert os.path.exists(ebas_local)
# else:
#     ebas_local=None


tda = AccessTestData()

TESTDATADIR = tda.basedir

OUTBASE = Path(TESTDATADIR).joinpath("testdata-minimal/obsdata/EBASMultiColumn")
SCRIPT_BASE_DIR = Path(TESTDATADIR).joinpath("testdata-minimal/scripts")

FILES_DEST = OUTBASE.joinpath("data")

UPDATE = True
UPDATE_EXISTING = False
SEARCH_PROBLEM_FILES = False
NAME = "EBASMC"

# if ebas_local is not None:
#     FILES_SRC = ebas_local
# else:
EBAS_BASE_DIR = "/lustre/storeA/project/aerocom/aerocom1/AEROCOM_OBSDATA/EBASMultiColumn/data/"
assert os.path.exists(EBAS_BASE_DIR)

JSON_FILE = SCRIPT_BASE_DIR.joinpath("ebas_files.json")

# ------------------------------------------------------------
# add some files with known problems
# at the moment these are static files whose names might change with every data set update
# from Nilu
# ------------------------------------------------------------
add_files = {
    # conco3 tower - 3 different measurement heights
    "o3_tower": "CZ0003R.20150101000000.20181107114213.uv_abs.ozone.air.1y.1h..CZ06L_uv_abs.lev2.nas",
    # conco3 Neg. meas periods
    "o3_neg_dt": "NZ0003G.20090110030000.20181130115605.uv_abs.ozone.air.9h.1h.US06L_Thermo_49C_LAU.US06L_AM.lev2.nas",
    # conco3 - Most common meas period is 150s and does not correspond to any of the supported base frequencies ['minutely', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'native']
    "o3_tstype": "LT0015R.20080101000000.20081231000000.uv_abs.ozone.air.15d.1h.LT01L_uv_abs_15.LT01L_uv_abs..nas",
    # concpm10 - could not resolve unique data column for concpm10 (EBAS varname: ['pm10_mass'])
    "pm10_colsel": "ID1013R.20180101000000.20200102000000.beta_gauge_particulate_sampler.pm10_mass.pm10.1y.1h.ID01L_MetOne_BAM1020..lev2.nas",
    # concpm10 Aliartos - Most common meas period is 172800s and does not correspond to any of the supported base frequencies ['minutely', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'native']
    "pm10_tstype": "GR0001R.20060119000000.20080120000000.b-attenuation.pm10_mass.pm10.11mo.1w.GR01L_b_att_01.GR01L_b-attenuation..nas",
}


def check_outdated(filedir):
    all_files = os.listdir(filedir)

    ok = True
    files_invalid = []
    files_valid = []

    with open(JSON_FILE) as f:
        data = simplejson.load(f, allow_nan=True)

    for var, stats in data.items():
        for stat, files in stats.items():
            for file in files:
                if file not in all_files:
                    print("OUTDATED", var, stat)
                    print(file)
                    files_invalid.append(file)
                    ok = False
                else:
                    files_valid.append(file)

    for key, file in add_files.items():
        if file not in all_files:
            print("OUTDATED (add_files)", key)
            print(file)
            files_invalid.append(file)
            ok = False
        else:
            files_valid.append(file)
    return ok, files_valid, files_invalid


def get_files_var_statlist(data, var, statlist, overlap_found):
    files = {}
    for stat in statlist:
        meta_match = data.find_station_meta_indices(stat)

        for i, idx in enumerate(meta_match):
            try:
                sd = data.to_station_data(idx, var)
            except pya.exceptions.DataCoverageError:
                continue
            if len(sd[var].dropna()) > 10:
                fname = sd.filename
                if ";" in fname:
                    raise ValueError("FATAL", var, stat)
                files[sd.station_name] = [sd.filename]
                break

        if not overlap_found:
            for left, right in zip(meta_match[:-1], meta_match[1:]):
                try:
                    sdl = data.to_station_data(left, var)
                    sdr = data.to_station_data(right, var)
                except pya.exceptions.DataCoverageError:
                    continue
                f0, f1 = sdl.filename, sdr.filename
                merged = sdl.merge_other(sdr, var)
                if var in merged.overlap:
                    files[sdl.station_name] = [f0, f1]
                    overlap_found = True
    return files, overlap_found


def get_files_var_statnum(data, var, statnum):
    files = {}
    numfound = 0
    allstats = data.unique_station_names
    for stat in allstats:
        if numfound == statnum:
            break
        meta_match = data.find_station_meta_indices(stat)

        for i, idx in enumerate(meta_match):
            try:
                sd = data.to_station_data(idx, var)
            except pya.exceptions.DataCoverageError:
                continue
            if len(sd[var].dropna()) > 10:
                fname = sd.filename
                if ";" in fname:
                    raise ValueError("FATAL", var, stat)
                files[sd.station_name] = [sd.filename]
                numfound += 1
                break
    return files


def main():
    # reader = pya.io.ReadUngridded(NAME, data_dir=EBAS_BASE_DIR)
    reader = pya.io.ReadUngridded(
        NAME,
    )
    r_lowlev = reader.get_lowlevel_reader(NAME)

    # r_lowlev._dataset_path = ebas_local

    # directory containing NASA Ames files
    FILEDIR_SRC = r_lowlev.file_dir
    print("checking for invalid files in static file list...")
    CURRENT_OK, files_valid, files_invalid = check_outdated(FILEDIR_SRC)
    print("done...")

    if not CURRENT_OK:
        print(f"outdated files: {files_invalid}")
        print("---------------------WARNING------------------------------")
        print(
            "Some files in current EBAS subset do not exist anymore in "
            "the most recent EBAS database, consider updating them"
        )
        print("---------------------WARNING------------------------------")
        raise Exception

    if not OUTBASE.exists():
        OUTBASE.mkdir()

    if not FILES_DEST.exists():
        FILES_DEST.mkdir()

    if (
        SEARCH_PROBLEM_FILES
    ):  # use this block to find interesting files, e.g. that throw certain exceptions
        var = "concpm10"
        f0 = 1100
        f1 = f0 + 1000

        reader = pya.io.ReadEbas()
        reader.read(var, first_file=f0, last_file=f1)
        for i, file in enumerate(reader.files_failed):
            try:
                data = reader.read_file(file, var)
            except Exception as e:
                print(i)
                print(os.path.basename(file))
                print(e)
                try:
                    fd = pya.io.EbasNasaAmesFile(file)
                    print(fd.station_name)
                except Exception as e:
                    print(e)

    else:
        infofile = JSON_FILE
        if os.path.exists(infofile):
            with open(infofile) as f:
                current_files = simplejson.load(f, allow_nan=True)
        else:
            current_files = {}

        VARS_STATFIX = ["sc550dryaer", "ac550aer", "conco3", "concpm10"]
        VARS_STATVAR_NUM = 2
        VARS_STATVAR = ["concno2", "vmrno2", "vmro3", "concpm25"]

        STATSFIX = ["Jungfraujoch", "Birkenes II", "*Kosetice*", "Troll"]

        all_files = {}
        overlap_found = False
        for var in VARS_STATFIX + VARS_STATVAR:
            if var in current_files and not UPDATE_EXISTING:
                all_files[var] = current_files[var]
                continue
            print(var)
            ropts = {}
            if var.startswith("conc") or var.startswith("vmr"):
                ropts["try_convert_vmr_conc"] = False
            print(f"reading var {var}")
            data = reader.read(NAME, var, **ropts)
            print("reading done")
            if var in VARS_STATFIX:
                files, overlap_found = get_files_var_statlist(data, var, STATSFIX, overlap_found)
            else:
                files = get_files_var_statnum(data, var, VARS_STATVAR_NUM)
            all_files[var] = files

        if not UPDATE:
            print("NOTHING WILL BE COPIED TO TEST DATA")
        else:
            src = Path(EBAS_BASE_DIR).joinpath("data")
            print(f"updating test data @ {r_lowlev.DATASET_PATH}")

            # copy revision file
            revision_file = os.path.join(r_lowlev.data_dir, r_lowlev.REVISION_FILE)
            outfile = OUTBASE.joinpath("Revision.txt")
            print(f"copying {revision_file} to {outfile}")
            shutil.copy(revision_file, outfile)

            # copy sqlite3 file
            db_file = r_lowlev.sqlite_database_file
            outfile = OUTBASE.joinpath(os.path.basename(db_file))
            print(f"copy {db_file} to {outfile}", db_file)
            shutil.copy(db_file, outfile)

            for var, finfo in all_files.items():
                print(var)
                for stat, files in finfo.items():
                    print(stat)
                    for file in files:
                        # print(file)
                        fp = src.joinpath(file)
                        dest = FILES_DEST.joinpath(file)
                        print(f"copying {fp} to {dest}")
                        shutil.copy(fp, dest)
                    print()

            for what, file in add_files.items():
                fp = src.joinpath(file)
                dest = FILES_DEST.joinpath(file)
                print(f"copying {fp} to {dest}")
                shutil.copy(fp, dest)

            print(f"re-writing {JSON_FILE}")
            with open(JSON_FILE, "w") as f:
                simplejson.dump(all_files, f, allow_nan=True)


if __name__ == "__main__":
    main()



from __future__ import annotations

import pandas as pd

from pyaerocom import UngriddedData
from pyaerocom.io import ReadPyaro


def test_testfile(pyaro_test_data_file):
    assert pyaro_test_data_file.exists()


def test_readpyaro(pyaro_testdata):
    rp = pyaro_testdata

    assert isinstance(rp, ReadPyaro)


def test_variables(pyaro_testdata):
    rp = pyaro_testdata
    variables = ["NOx", "concso4", "od550aer"]

    assert rp.PROVIDES_VARIABLES == variables
    assert rp.DEFAULT_VARS == variables


####################################################
#   Tests for the helper class PyaroToUngriddedData
####################################################


def test_pyarotoungriddeddata_reading(pyaro_testdata):
    from math import ceil

    obj = pyaro_testdata.converter
    data = obj.read()
    assert isinstance(data, UngriddedData)

    # Checks is data is empty
    assert not data.is_empty
    assert len(data.unique_station_names) == 2

    # Tests the found stations
    all_stations = data.to_station_data_all("concso4", ts_type_preferred="daily")

    assert all_stations["stats"][0]["ts_type"] in ["hourly", "3daily", "2hourly", "2daily"]
    assert all_stations["stats"][0]["country"] == "NO"

    # Tests the dates
    start = pd.to_datetime("01.01.2010", dayfirst=True)
    end = pd.to_datetime("31.12.2010", dayfirst=True)
    dates = pd.date_range(start, end, freq="D")
    assert len(all_stations["stats"][0].dtime) == ceil(len(dates) / 2)


def test_pyarotoungriddeddata_reading_kwargs(pyaro_testdata_kwargs):
    obj = pyaro_testdata_kwargs.converter
    data = obj.read()
    assert isinstance(data, UngriddedData)

    # Checks if stations have correct countries
    all_stations = data.to_station_data_all("concso4")
    countries = ["NO", "GB"]
    assert all_stations["stats"][1]["country"].strip() == countries[0]
    assert all_stations["stats"][0]["country"].strip() == countries[1]


def test_pyarotoungriddeddata_reading_extra_metadata(pyaro_testdata_kwargs):
    obj = pyaro_testdata_kwargs.converter
    data = obj.read()
    assert isinstance(data, UngriddedData)

    # Checks if stations have correct countries
    all_stations = data.to_station_data_all("concso4", add_meta_keys=["area_classification"])
    area_type = ["Rural", "Urban"]
    assert all_stations["stats"][1]["area_classification"].strip() == area_type[0]
    assert all_stations["stats"][0]["area_classification"].strip() == area_type[1]


def test_pyarotoungriddeddata_stations(pyaro_testdata):
    obj = pyaro_testdata.converter

    assert len(obj.get_stations()) == 2


def test_pyarotoungriddeddata_variables(pyaro_testdata):
    obj = pyaro_testdata.converter

    assert obj.get_variables() == pyaro_testdata.PROVIDES_VARIABLES



from pathlib import Path

import pytest

from pyaerocom import UngriddedData
from pyaerocom.io import ReadAeronetSunV3
from pyaerocom.io.cachehandler_ungridded import CacheHandlerUngridded
from tests.conftest import lustre_avail


@pytest.fixture(scope="module")
def cache_handler():
    return CacheHandlerUngridded()


@lustre_avail
def test_cache_dir(cache_handler: CacheHandlerUngridded):
    cache_path = Path(cache_handler.cache_dir)
    comps = cache_path.parts
    assert comps[-2] == "_cache"
    assert comps[-3] == "MyPyaerocom"


def test_reload_custom(
    cache_handler: CacheHandlerUngridded, aeronetsunv3lev2_subset: UngriddedData, tmp_path: Path
):
    path = tmp_path / "test_manual_caching.pkl"
    cache_handler.write(aeronetsunv3lev2_subset, var_or_file_name=path.name, cache_dir=path.parent)
    assert path.exists()
    cache_handler.check_and_load(var_or_file_name=path.name, cache_dir=path.parent)
    assert cache_handler.loaded_data[path.name].shape == aeronetsunv3lev2_subset.shape


@pytest.mark.dependency
def test_reload(
    cache_handler: CacheHandlerUngridded,
    aeronetsunv3lev2_subset: UngriddedData,
    aeronet_sun_subset_reader: ReadAeronetSunV3,
):
    cache_handler.reader = aeronet_sun_subset_reader
    cache_handler.write(aeronetsunv3lev2_subset, var_or_file_name="od550aer")
    assert Path(cache_handler.file_path("od550aer")).exists()

    cache_handler.check_and_load(var_or_file_name="od550aer")
    subset = aeronetsunv3lev2_subset.extract_var("od550aer")
    assert "od550aer" in cache_handler.loaded_data

    reloaded = cache_handler.loaded_data["od550aer"]
    assert isinstance(reloaded, UngriddedData)
    assert reloaded.shape == subset.shape



#!/usr/bin/env python3
"""
Create minimal testdataset for GHOST reader

Created on Fri Feb 26 09:17:09 2021

@author: jonasg
"""

import os

import matplotlib.pyplot as plt

plt.close("all")
import xarray as xr

import pyaerocom as pya

path_in = os.path.join(pya.const.OUTPUTDIR, "data/obsdata/GHOST/data")

path_out = os.path.join(pya.const.OUTPUTDIR, "testdata-minimal/obsdata/GHOST/data")

assert os.path.exists(path_in)
assert os.path.exists(path_out)

datasets = ["EEA_AQ_eReporting", "EBAS"]

freqs = ["hourly", "daily"]

varis = ["pm10", "sconco3"]
datesfiles = ["201810", "201911", "201912"]


def filename(var: str, date: str) -> str:
    return f"{var}_{date}.nc"


files_out = []
for dsname in datasets:
    for freq in freqs:
        indir = os.path.join(path_in, dsname, freq)
        assert os.path.exists(indir)
        outdir = os.path.join(path_out, dsname, freq)
        os.makedirs(outdir, exist_ok=True)
        assert os.path.exists(outdir)
        for var in varis:
            if var == "pm10":
                dates = datesfiles
                numst = 3

                numts = None if freq == "daily" else 3

            else:
                dates = [datesfiles[0]]
                numst = 1
                numts = 3
            for date in dates:
                dir_in = os.path.join(indir, var)
                dir_out = os.path.join(outdir, var)
                os.makedirs(dir_out, exist_ok=True)
                assert os.path.exists(dir_in)
                fname = filename(var, date)
                file_in = os.path.join(dir_in, fname)
                file_out = os.path.join(dir_out, fname)
                print(file_in)
                print(file_out)
                assert os.path.exists(file_in)

                with xr.open_dataset(file_in) as ds:
                    subset = ds.isel(station=slice(0, numst))
                    if numts is not None:
                        subset = subset.isel(time=slice(0, numts))

                    subset.to_netcdf(file_out)
                print("Saved")



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from matplotlib.axes import Axes
from xarray import DataArray

from pyaerocom.exceptions import (
    CoordinateError,
    DataUnitError,
    MetaDataError,
    UnitConversionError,
    VarNotAvailableError,
)
from pyaerocom.io import ReadEarlinet
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from tests.conftest import TEST_RTOL
from tests.fixtures.stations import FAKE_STATION_DATA


def get_earlinet_data(var_name):
    data = ReadEarlinet("Earlinet-test").read(vars_to_retrieve=var_name)
    stats = data.to_station_data_all()["stats"]
    assert len(stats) == 1
    return stats[0]


stat1 = FAKE_STATION_DATA["station_data1"]
stat2 = FAKE_STATION_DATA["station_data2"]


def test_StationData_copy():
    cp = stat1.copy()
    for key, val in stat1.items():
        assert key in cp
        if isinstance(val, np.ndarray):
            assert np.all(val == cp[key])
        else:
            assert val == cp[key]


stat3 = stat2.copy()
stat3["concno2"] = np.ones(10)
stat3["var_info"]["concno2"] = dict(unit="ug m-3")
stat3["concso4"] = np.ones(10)
stat3["var_info"]["concso4"] = dict(units="1")
stat3["ts_type"] = None
stat3["station_coords"]["latitude"] = "blaaa"
stat4 = stat2.copy()
stat4["longitude"] = "42"

ec_earlinet = get_earlinet_data("ec355aer")


def test_StationData_default_vert_grid():
    grid = stat1.default_vert_grid
    assert grid.mean() == 7375  # m
    step = np.unique(np.diff(grid))
    assert len(step) == 1
    assert step[0] == 250


def test_StationData_vars_available():
    assert stat1.vars_available == ["ec550aer", "od550aer"]


def test_StationData_has_var():
    assert stat1.has_var("od550aer")
    assert stat2.has_var("conco3")
    assert not stat2.has_var("abs550aer")

    copy = stat2.copy()
    copy["abs550aer"] = np.ones(10)
    assert copy.has_var("abs550aer")


def test_StationData_get_unit():
    assert stat1.get_unit("ec550aer") == "m-1"


@pytest.mark.parametrize(
    "stat,var_name,error",
    [
        pytest.param(
            stat1,
            "concco",
            "Could not access variable metadata dict for concco.",
            id="no metadata",
        ),
        pytest.param(
            stat3,
            "concno2",
            "Failed to access units attribute for variable concno2. "
            'Corresponding var_info dict contains attr. "unit", which is deprecated, please check corresponding reading routine. ',
            id="wrong metadata",
        ),
    ],
)
def test_StationData_get_unit_error(stat: StationData, var_name: str, error: str):
    with pytest.raises(MetaDataError) as e:
        stat.get_unit(var_name)
    assert str(e.value) == error


def test_StationData_units():
    assert stat1.units == {"ec550aer": "m-1", "od550aer": "1"}


def test_StationData_check_var_unit_aerocom():
    stat = stat1.copy()
    assert stat.get_unit("ec550aer") == "m-1"

    stat.check_var_unit_aerocom("ec550aer")
    assert stat.get_unit("ec550aer") == "1/km"


@pytest.mark.parametrize(
    "stat,var_name,exception,error",
    [
        pytest.param(
            stat1.copy(),
            "concco",
            MetaDataError,
            "Could not access variable metadata dict for concco.",
            id="var not found",
        ),
        pytest.param(
            stat3.copy(),
            "concso4",
            UnitConversionError,
            "failed to convert unit from 1 to ug m-3",
            id="can not convert",
        ),
    ],
)
def test_StationData_check_var_unit_aerocom_error(
    stat: StationData, var_name: str, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        stat.check_var_unit_aerocom(var_name)
    assert str(e.value) == error


def test_StationData_check_unit():
    stat1.check_unit("ec550aer", "m-1")


def test_StationData_check_unit_error():
    with pytest.raises(DataUnitError) as e:
        stat1.check_unit("ec550aer", None)
    assert str(e.value) == "Invalid unit m-1 (expected 1/km)"


def test_StationData_convert_unit():
    stat1.convert_unit("ec550aer", "1/Gm")


def test_StationData_convert_unit_error():
    with pytest.raises(UnitConversionError) as e:
        stat3.convert_unit("concso4", "kg m-3")
    assert str(e.value) == "failed to convert unit from 1 to kg m-3"


def test_StationData_dist_other():
    dist = stat1.dist_other(stat2)
    assert dist == pytest.approx(1.11, abs=0.1)


@pytest.mark.parametrize(
    "stat,other,tol_km,result",
    [
        (stat1, stat1, None, True),
        (stat1, stat2, 1, False),
        (stat1, stat2, 2, True),
    ],
)
def test_StationData_same_coords(
    stat: StationData, other: StationData, tol_km: float | None, result: bool
):
    assert stat.same_coords(other, tol_km) == result


@pytest.mark.parametrize(
    "stat,force_single_value,dtype",
    [
        (StationData(), False, None),
        (stat1, False, None),
        (stat1, True, float),
        (stat2, True, float),
        (
            StationData(station_coords=dict(latitude=42, longitude=42, altitude=42)),
            False,
            None,
        ),
    ],
)
def test_StationData_get_station_coords(
    stat: StationData, force_single_value: bool, dtype: type | None
):
    coords = stat.get_station_coords(force_single_value=force_single_value)
    assert isinstance(coords, dict)
    if dtype is not None:
        for val in coords.values():
            assert isinstance(val, dtype)


@pytest.mark.parametrize(
    "stat,exception,error",
    [
        pytest.param(
            stat3,
            MetaDataError,
            "Station coordinate latitude must be numeric. Got: blaaa",
            id="wrong station coordinate",
        ),
        pytest.param(
            stat4,
            AttributeError,
            "Invalid value encountered for coord longitude, need float, int, list or ndarray, got <class 'str'>",
            id="wrong station latitude",
        ),
        pytest.param(
            StationData(latitude=[42, 44, 32], longitude=42, altitude=42),
            ValueError,
            "meas point coordinate arrays of latitude vary too much to reduce them to a single coordinate. "
            "Order of difference in latitude is 12 and maximum allowed is 0.05.",
            id="tolerance",
        ),
    ],
)
def test_StationData_get_station_coords_error(
    stat: StationData, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        stat.get_station_coords(force_single_value=True)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "add_none_vals,add_meta_keys,num",
    [
        (False, None, 14),
        (False, "blaaa", 14),
        (False, ["random_key1", "random_key2"], 16),
        (False, ["random_key4"], 15),
        (True, ["random_key4"], 23),
    ],
)
@pytest.mark.parametrize("stat,force_single_value,quality_check", [(stat1, True, True)])
def test_StationData_get_meta(
    stat: StationData,
    force_single_value: bool,
    quality_check: bool,
    add_none_vals: bool,
    add_meta_keys: list[str],
    num: int,
):
    meta = stat.copy().get_meta(
        force_single_value, quality_check, add_none_vals, add_meta_keys=add_meta_keys
    )
    assert isinstance(meta, dict)
    assert len(meta) == num


def test_StationData_get_meta_error():
    with pytest.raises(MetaDataError) as e:
        stat1.copy().get_meta(add_none_vals=False, add_meta_keys=["random_key3"])
    assert str(e.value) == "Inconsistencies in meta parameter random_key3"


@pytest.mark.parametrize("key", ["station_name", "framework", "longitude"])
@pytest.mark.parametrize("stat", [stat1.copy()])
def test_StationData__check_meta_item(stat: StationData, key: str):
    stat._check_meta_item(key)


@pytest.mark.parametrize(
    "other,coord_tol_km,inplace,add_meta_keys",
    [
        (stat1, 0.1, True, None),
        (stat2, 50, True, None),
        (stat2, 50, True, ["random_key1"]),
        (stat2, 50, False, ["random_key1"]),
    ],
)
@pytest.mark.parametrize("stat,check_coords,raise_on_error", [(stat1.copy(), True, True)])
def test_StationData_merge_meta_same_station(
    stat: StationData,
    other: StationData,
    coord_tol_km: float,
    check_coords: bool,
    inplace: bool,
    add_meta_keys: list[str],
    raise_on_error: bool,
):
    _stat = stat.merge_meta_same_station(
        other, coord_tol_km, check_coords, inplace, add_meta_keys, raise_on_error
    )
    assert isinstance(_stat, StationData)
    assert (_stat is stat) == inplace


def test_StationData_merge_meta_same_station_error():
    with pytest.raises(CoordinateError, match="differ by more than 0.001 km."):
        stat1.merge_meta_same_station(
            stat2, coord_tol_km=0.001, check_coords=True, inplace=False, raise_on_error=True
        )


@pytest.mark.parametrize("stat", [stat1.copy(), stat2.copy()])
@pytest.mark.parametrize("other", [stat1, stat2])
def test_StationData_merge_varinfo(stat: StationData, other: StationData):
    assert stat.merge_varinfo(other, "ec550aer") is stat


@pytest.mark.parametrize("stat,other", [(stat1.copy(), stat2), (stat2.copy(), stat1)])
def test_StationData_merge_varinfo_error(stat: StationData, other: StationData):
    with pytest.raises(MetaDataError) as e:
        stat.merge_varinfo(other, "conco3")
    assert str(e.value) == "No variable meta information available for conco3"


@pytest.mark.parametrize(
    "stat,var_name,result",
    [
        (stat1, "od550aer", False),
        (stat2, "od550aer", False),
    ],
)
def test_StationData_check_if_3d(stat: StationData, var_name: str, result: bool):
    assert stat.check_if_3d(var_name) == result


def test_StationData__check_ts_types_for_merge():
    assert stat1.copy()._check_ts_types_for_merge(stat2, "ec550aer") == "monthly"


@pytest.mark.parametrize(
    "low,high,mean",
    [
        (None, None, 1),
        (10, 20, np.nan),
    ],
)
@pytest.mark.parametrize("stat,var_name,check_unit", [(stat1.copy(), "od550aer", True)])
@pytest.mark.filterwarnings("ignore:Mean of empty slice:RuntimeWarning")
def test_StationData_remove_outliers(
    stat: StationData,
    var_name: str,
    low: float | None,
    high: float | None,
    check_unit: bool,
    mean: float,
):
    stat.remove_outliers(var_name, low, high, check_unit)
    avg = np.nanmean(stat[var_name])
    assert avg == pytest.approx(mean, rel=TEST_RTOL, nan_ok=True)


def test_StationData_calc_climatology(aeronetsunv3lev2_subset: UngriddedData):
    site = aeronetsunv3lev2_subset.to_station_data(6, vars_to_convert="od550aer")
    clim = site.calc_climatology("od550aer")
    assert clim is not site
    assert isinstance(clim, StationData)
    mean = np.nanmean(clim.od550aer)  # type:ignore[attr-defined]
    assert mean == pytest.approx(0.44, abs=0.01)


def test_StationData_remove_variable():
    stat = stat1.copy()

    var = "ec550aer"
    assert var in stat
    assert var in stat.var_info

    stat.remove_variable(var)
    assert var not in stat
    assert var not in stat.var_info


def test_StationData_remove_variable_error():
    with pytest.raises(VarNotAvailableError) as e:
        stat1.remove_variable("concco")
    assert str(e.value) == "No such variable in StationData: concco"


def test_StationData_select_altitude_DataArray():
    selection = ec_earlinet.select_altitude("ec355aer", (1000, 2000))
    assert isinstance(selection, DataArray) or isinstance(selection, pd.Series)
    assert selection.shape == (16,)


def test_StationData_select_altitude_DataArray_error():
    with pytest.raises(NotImplementedError) as e:
        ec_earlinet.select_altitude("ec355aer", 1000)
    assert str(e.value) == "So far only a range (low, high) is supported for altitude extraction."


def test_StationData_select_altitude_Series():
    series = stat1.select_altitude("od550aer", (100, 301))
    assert isinstance(series, pd.Series)


@pytest.mark.parametrize(
    "stat,altitudes,exception,error",
    [
        (stat1, (100, 101), ValueError, "no data in specified altitude range"),
        (stat2, (100, 1000), AttributeError, "need 1D altitude array"),
    ],
)
def test_StationData_select_altitude_Series_error(
    stat: StationData, altitudes: tuple[int, int], exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        stat.select_altitude("od550aer", altitudes)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "stat,var_name,kwargs",
    [
        (stat1, "od550aer", dict()),
    ],
)
def test_StationData_to_timeseries(stat: StationData, var_name: str, kwargs: dict):
    series = stat.to_timeseries(var_name, **kwargs)
    assert isinstance(series, pd.Series)


def test_StationData_plot_timeseries():
    ax = stat1.plot_timeseries(var_name="od550aer")
    assert isinstance(ax, Axes)


def test_StationData___str__():
    assert isinstance(str(stat1), str)



"""
Helper methods for access of and working with land/sea masks. pyaerocom
provides automatic access to HTAP land sea masks from this URL:

https://pyaerocom.met.no/pyaerocom-suppl

Filtering by these masks is implemented in :class:`Filter` and all relevant
data classes (i.e. :class:`GriddedData`, :class:`UngriddedData`,
:class:`ColocatedData`).
"""

import glob
import logging
import os

import numpy as np
import requests
import xarray as xr
from iris import load_cube

from pyaerocom import const
from pyaerocom.exceptions import DataRetrievalError
from pyaerocom.helpers import numpy_to_cube

logger = logging.getLogger(__name__)


def available_htap_masks():
    """
    List of HTAP mask names

    Returns
    ----------
    list
        Returns a list of available htap region masks.
    """
    return const.HTAP_REGIONS


def download_htap_masks(regions_to_download=None):
    """Download HTAP mask

    URL: https://pyaerocom.met.no/pyaerocom-suppl.

    Parameters
    -----------
    regions_to_download : list
        List containing the regions to download.

    Returns
    -------
    list
        List of file paths that point to the mask files that were successfully
        downloaded

    Raises
    ------
    ValueError
        if one of the input regions does not exist
    DataRetrievalError
        if download fails for one of the input regions
    """

    if regions_to_download is None:
        regions_to_download = const.HTAP_REGIONS
    elif isinstance(regions_to_download, str):
        regions_to_download = [regions_to_download]
    elif not isinstance(regions_to_download, list):
        raise ValueError("Invalid input for regions_to_download, need list or str")

    path_out = const.FILTERMASKKDIR
    base_url = const.URL_HTAP_MASKS

    paths = []
    for region in regions_to_download:
        if region not in const.HTAP_REGIONS:
            raise ValueError(f"No such HTAP region {region}")
        elif region == "EAS":
            filename = f"{region}htap.nc"
            file_out = os.path.join(path_out, f"{region}htap.0.1x0.1deg.nc")
        else:
            filename = f"{region}htap.0.1x0.1deg.nc"
            file_out = os.path.join(path_out, filename)

        url = os.path.join(base_url, filename)

        try:
            r = requests.get(url)
            open(file_out, "wb").write(r.content)
            paths.append(file_out)
        except Exception as e:
            raise DataRetrievalError(f"Failed to download HTAP mask {region}. Reason {repr(e)}")
    return paths


def get_htap_mask_files(*region_ids):
    """Get file paths to input HTAP regions

    Parameters
    ----------
    *region_ids
        ID's of regions for which mask files are supposed to be retrieved

    Returns
    -------
    list
        list of file paths for each input region

    Raises
    ------
    FileNotFoundError
        if default local directory for storage of HTAP masks does not exist
    NameError
        if multiple mask files are found for the same region
    """
    mask_dir = const.FILTERMASKKDIR
    if not os.path.exists(mask_dir):
        raise FileNotFoundError("HTAP mask directory does not exist")
    out = []
    for region in region_ids:
        if region not in const.HTAP_REGIONS:
            raise ValueError(f"No such HTAP region {region}")
        files = glob.glob(os.path.join(mask_dir, f"{region}*.nc"))
        if len(files) != 1:
            if len(files) == 0:
                logger.info(f"Downloading HTAP mask {region}")
                files = download_htap_masks(region)
            elif len(files) > 1:
                raise NameError(f"Found multiple masks for region {region}")
        out.append(files[0])
    return out


def load_region_mask_xr(*regions):
    """Load boolean mask for input regions (as xarray.DataArray)

    Parameters
    -----------
    *regions
        regions that are supposed to be loaded and merged (just use string,
        no list or similar)

    Returns
    ---------
    xarray.DataArray
        boolean mask for input region(s)
    """
    masks = None
    for i, fil in enumerate(get_htap_mask_files(*regions)):
        r = regions[i]
        if i == 0:
            masks = xr.load_dataset(fil)[r + "htap"]
            name = r
        else:
            masks += xr.load_dataset(fil)[r + "htap"]
            name += f"-{r}"
    if masks is not None:
        mask = masks.where(masks < 1, 1)
    mask["name"] = name
    mask.attrs["long_name"] = name
    mask = mask.rename({"lat": "latitude", "long": "longitude"})
    return mask


def load_region_mask_iris(*regions):
    """Loads regional mask to iris.

    Parameters
    -----------
    region_id : str
        Chosen region.

    Returns
    ---------
    iris.cube.Cube
        cube representing merged mask from input regions
    """
    cubes = []
    names = []
    for i, fil in enumerate(get_htap_mask_files(*regions)):
        names.append(regions[i])
        cubes.append(load_cube(fil))
    if len(cubes) == 1:
        out = cubes[0]
    else:
        merged_np = np.max([x.data for x in cubes], axis=0)
        out = numpy_to_cube(merged_np, dims=(cubes[0].coords()[0], cubes[0].coords()[1]))
        out.units = cubes[0].units
    name = "-".join(names)
    out.var_name = name
    # out.attributes['long_name'] = name
    return out


def get_mask_value(lat, lon, mask):
    """Get value of mask at input lat / lon position

    Parameters
    ----------
    lat : float
        latitute
    lon : float
        longitude
    mask : xarray.DataArray
        data array

    Returns
    -------
    float
        neirest neigbhour mask value to input lat lon
    """
    if not isinstance(mask, xr.DataArray):
        raise ValueError(f"Invalid input for mask: need DataArray, got {type(mask)}")
    return float(mask.sel(latitude=lat, longitude=lon, method="nearest"))


def check_all_htap_available():
    """
    Check for missing HTAP masks on local computer and download
    """
    return get_htap_mask_files(*available_htap_masks())


def get_lat_lon_range_mask_region(mask, latdim_name=None, londim_name=None):
    """
    Get outer lat/lon rectangle of a binary mask

    Parameters
    ----------
    mask : xr.DataArray
        binary mask
    latdim_name : str, optional
        Name of latitude dimension. The default is None, in which case lat is
        assumed.
    londim_name : str, optional
        Name of longitude dimension. The default is None, in which case long is
        assumed.

    Returns
    -------
    dict
        dictionary containing lat and lon ranges of the mask.

    """
    if latdim_name is None:  # htap
        latdim_name = "lat"
    if londim_name is None:
        londim_name = "long"  # htap
    assert isinstance(mask, xr.DataArray)
    assert mask.dims == (latdim_name, londim_name)

    data = mask.data
    lats = mask.latitude.data
    lons = mask.longitude.data

    lonmask = np.where(data.any(axis=0))[0]  # flatten latitude dimenstion
    firstidx, lastidx = lonmask.min(), lonmask.max()
    lonr = sorted([lons[firstidx], lons[lastidx]])

    latmask = np.where(data.any(axis=1))[0]  # flatten latitude dimenstion
    firstidx, lastidx = latmask.min(), latmask.max()
    latr = sorted([lats[firstidx], lats[lastidx]])

    return dict(lat_range=latr, lon_range=lonr)



"""
Pyearocom module for reading and processing of EBAS NASA Ames files

For details on the file format see `here <https://ebas-submit.nilu.no/
Submit-Data/Getting-started>`__
"""

import csv
import logging
import os
from datetime import datetime
from io import StringIO

import numpy as np

from pyaerocom import const
from pyaerocom._lowlevel_helpers import dict_to_str, str_underline
from pyaerocom.exceptions import NasaAmesReadError, TimeZoneError

logger = logging.getLogger(__name__)


class EbasColDef(dict):
    """Dict-like object for EBAS NASA Ames column definitions

    Note
    ----
    The meta attribute name 'unit' can also be accessed using the CF attr name
    'units'

    Attributes
    ----------
    name : str
        column name
    unit : str
        unit of data in column (if applicable)
    is_var : bool
        True if column corresponds to variable data, False if not
    is_flag : bool
        True, if column corresponds to Flag column, False if not
    flag_col : int
        column number of flag column that corresponds to this data column (only
        relevant if :attr:`is_var` is True)

    Parameters
    ----------
    name : str
        column name
    is_var : bool
        True if column corresponds to variable data, False if not
    is_flag : bool
        True, if column corresponds to Flag column, False if not
    unit : :obj:`str`, optional
        unit of data in column (if applicable)
    flag_col : :obj:`str`, optional
        ``name`` of flag column that corresponds to this data colum (only
        relevant if :attr:`is_var` is True)
    """

    def __init__(self, name, is_var, is_flag, unit="1"):
        self.name = name
        self.unit = unit
        self.is_var = is_var
        self.is_flag = is_flag
        self.flag_col = None

    def get_wavelength_nm(self):
        """Try to access wavelength information in nm (as float)"""
        if "wavelength" not in self:
            raise KeyError(f"Column variable {self.name} does not contain wavelength information")
        elif "nm" not in self.wavelength:
            raise NotImplementedError("Wavelength definition is not in nm")
        return float(self.wavelength.split("nm")[0].strip())

    def to_dict(self, ignore_keys=["is_var", "is_flag", "flag_col", "wavelength_nm"]):
        d = {}
        for k, v in self.items():
            if k in ignore_keys:
                continue
            elif k == "unit":
                k = "units"
            d[k] = v
        return d

    def __getitem__(self, key):  # add support for units attr (CF standard name)
        if key == "units":
            key = "unit"
        return super().__getitem__(key)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, val):
        self[key] = val

    def __repr__(self):
        return str(self)

    def __str__(self):
        colattrs = ""
        for key, val in self.items():
            colattrs += f"{key}: {val}, "
        return colattrs[:-2]


def _readline_ref_and_revision(line):
    spl = line.strip().split()
    basedate = np.datetime64(datetime.strptime(f"{spl[0]}{spl[1]}{spl[2]}", "%Y%m%d"), "s")
    rev = np.datetime64(datetime.strptime(f"{spl[3]}{spl[4]}{spl[5]}", "%Y%m%d"), "s")
    return [basedate, rev]


class NasaAmesHeader:
    """Header class for Ebas NASA Ames file

    Note
    ----
    Is used in :class:`EbasNasaAmesFile` and should not be used directly.

    Attributes
    -----------

    """

    _NUM_FIXLINES = 13
    _HEAD_ROWS_MANDATORY = [0, 5, 8, 9, 10, 11]

    # conversion methods for first 13 header lines of
    CONV_STR = lambda l: str(l.strip())  # noqa: E731,E741
    CONV_PI = lambda l: "; ".join([x.strip() for x in l.split(";")])  # noqa: E731,E741
    CONV_MULTIINT = lambda l: [int(x) for x in l.strip().split()]  # noqa: E731,E741
    CONV_MULTIFLOAT = lambda l: [float(x) for x in l.strip().split()]  # noqa: E731,E741
    CONV_INT = lambda l: int(l.strip())  # noqa: E731,E741
    CONV_FLOAT = lambda l: float(l.strip())  # noqa: E731,E741
    _STARTDATE_FMT = "%Y%m%d%H%M%S"

    # the parameters that are always in the beginning of the file
    _H_FIXLINES_YIELD = [
        ["num_head_lines", "num_head_fmt"],  # line 1
        "data_originator",  # line 2
        "sponsor_organisation",  # 3
        "submitter",  # 4
        "project_association",  # 5
        ["vol_num", "vol_totnum"],  # 6
        ["ref_date", "revision_date"],  # 7
        "freq",  # 8
        "descr_time_unit",  # 9
        "num_cols_dependent",  # 10
        "mul_factors",  # 11
        "vals_invalid",  # 12
        "descr_first_col",  # 13
    ]
    # conversion methods are defined above
    _H_FIXLINES_CONV = [
        CONV_MULTIINT,  # 1 -> yields 2
        CONV_PI,  # 2
        CONV_STR,  # 3
        CONV_STR,  # 4
        CONV_STR,  # 5
        CONV_MULTIINT,  # 6
        _readline_ref_and_revision,  # 7 lambda l : [x.strip() for x in l.strip().split("     ")], #7
        CONV_FLOAT,  # 8
        CONV_STR,  # 9
        CONV_INT,  # 10
        CONV_MULTIFLOAT,  # 11
        CONV_MULTIFLOAT,  # 12
        CONV_STR,
    ]  # 13

    def __init__(self, **kwargs):
        self._head_fix = dict(
            num_head_lines=np.nan,
            num_head_fmt=np.nan,
            data_originator="",
            sponsor_organisation="",
            submitter="",
            project_association="",
            vol_num=np.nan,
            vol_totnum=np.nan,
            ref_date=np.nan,
            revision_date=np.nan,
            freq=np.nan,
            descr_time_unit="",
            num_cols_dependent=np.nan,
            mul_factors=[],
            vals_invalid=[],
            descr_first_col="",
        )
        self._var_defs = []
        self._meta = {}
        self.update(**kwargs)

    @property
    def head_fix(self):
        """Dictionary containing fixed header info (that is always available)"""
        return self._head_fix

    @property
    def var_defs(self):
        """List containing column variable definitions

        List index is column index in file and value is instance of
        :class:`EbasColDef`
        """
        return self._var_defs

    @property
    def meta(self):
        """Meta data dictionary (specific for this file)"""
        return self._meta

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    def __getattr__(self, key):
        if key in self._head_fix:
            return self._head_fix[key]
        elif key in self._meta:
            return self._meta[key]
        else:
            raise AttributeError(f"Invalid attribute: {key}")

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, val):
        if key in self._head_fix:
            self._head_fix[key] = val
        else:
            self._meta[key] = val

    def __str__(self):
        head = f"{type(self).__name__}"
        s = f"{head}\n{len(head)*'-'}\n"
        s += dict_to_str(self._head_fix)
        s += f"\n\n{str_underline('Column variable definitions', indent=3)}"
        for item in self._var_defs:
            s += f"\n   {repr(item)}"
        s += f"\n\n{str_underline('EBAS meta data', indent=3)}"
        s += dict_to_str(self.meta)

        return s


class EbasFlagCol:
    """Simple helper class to decode and interpret EBAS flag columns

    Attributes
    ----------
    raw_data : ndarray
        raw flag column (containing X-digit floating point numbers)

    """

    def __init__(self, raw_data, interpret_on_init=True):
        self.raw_data = raw_data

        self._decoded = None
        self._valid = None

        if interpret_on_init:
            self.decode()

    @property
    def FLAG_INFO(self):
        """Detailed information about EBAS flag definitions"""
        return const.ebas_flag_info

    @property
    def decoded(self):
        """Nx3 numpy array containing decoded flag columns"""
        if self._decoded is None:
            self.decode()
        return self._decoded

    @property
    def valid(self):
        """Boolean array specifying valid and invalid measurements"""
        if self._valid is None:
            self.decode()
        return self._valid

    def decode(self):
        """Decode raw flag column"""
        flags = np.zeros((len(self.raw_data), 3)).astype(int)
        mask = self.raw_data.astype(
            bool
        )  # rmooves all points that are 0, i.e. that contain no flag (valid measurements)
        valid = np.ones_like(self.raw_data).astype(bool)
        not_ok = self.raw_data[mask]
        if len(not_ok) > 0:
            _decoded = []
            _valid = []
            for flag in not_ok:
                item = f"{flag:.9f}".split(".")[1]
                vals = [int(item[:3]), int(item[3:6]), int(item[6:9])]
                _invalid = False
                for val in vals:
                    if val == 100:
                        _invalid = False
                        break  # since all other flags are irrelevant if 100 is flagged
                    elif (
                        val in self.FLAG_INFO["valid"] and not self.FLAG_INFO["valid"][val]
                    ):  # is invalid
                        _invalid = True

                _decoded.append(vals)
                _valid.append(not _invalid)

            flags[mask] = np.asarray(_decoded)
            valid[mask] = _valid

        self._valid = valid
        self._decoded = flags


class EbasNasaAmesFile(NasaAmesHeader):
    """EBAS NASA Ames file interface

    Class interface for reading and processing of EBAS NASA Ames file

    Attributes
    ----------
    time_stamps : ndarray
        array containing datetime64 objects with timestamps
    flags : dict
        dictionary containing :class:`EbasFlagCol` objects for each column
        containing flags

    Parameters
    ----------
    file : :obj:`str`, optional
        EBAS NASA Ames file. if valid file path, then the file is read on
        init (please note following options for import)
    only_head : bool
        read only file header
    replace_invalid_nan : bool
        replace all invalid values in the table by NaNs. The invalid values for
        each dependent data column are identified based on the information in
        the file header.
    convert_timestamps : bool
        compute array of numpy datetime64 timestamps from numeric timestamps
        in data
    evaluate_flags : bool
        if True, all flags in all flag columns are decoded from floating
        point representation to 3 integers, e.g.
        0.111222333 -> 111 222 333
    quality_check : bool
        perform quality check after import (for details see
        :func:`_quality_check`)
    **kwargs
        optional input args that are passed to init of :class:`NasaAmesHeader`
        base class

    """

    TIMEUNIT2SECFAC = dict(days=3600 * 24, Days=3600 * 24)

    ERR_LOW_STATS = "percentile:15.87"
    ERR_HIGH_STATS = "percentile:84.13"

    def __init__(
        self,
        file=None,
        only_head=False,
        replace_invalid_nan=True,
        convert_timestamps=True,
        evaluate_flags=False,
        quality_check=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._data_header = []  # Header line of data block
        self._data = []  # data block

        self.time_stamps = None
        self.start_meas = None
        self.stop_meas = None

        self.flag_col_info = {}

        self.file = None

        if file is not None:
            if not os.path.exists(file):
                raise OSError(f"File {file} does not exists")
            self.read_file(
                file,
                only_head,
                replace_invalid_nan,
                convert_timestamps,
                evaluate_flags,
                quality_check,
            )

    @property
    def data(self):
        """2D numpy array containing data table"""
        return self._data

    @property
    def data_header(self):
        return self._data_header

    @property
    def shape(self):
        """Shape of data array"""
        return self.data.shape

    @property
    def col_num(self):
        """Number of columns in table"""
        return self.num_cols_dependent + 1

    @property
    def col_names(self):
        """Column names of table"""
        # cols = [x["name"] for x in self.var_defs]
        return [x["name"] for x in self.var_defs]

    @property
    def col_names_vars(self):
        """Names of all columns that are flagged as variables"""
        return [x.name for x in self.var_defs if x.is_var]

    @property
    def col_nums_vars(self):
        """Column index number of all variables"""
        return [i for (i, item) in enumerate(self.var_defs) if item.is_var]

    @property
    def base_date(self):
        """Base date of data as numpy.datetime64[s]"""
        if "timezone" not in self.meta:
            raise AttributeError(
                "Fatal: could not infer base date. Timezone is not available in file header"
            )
        if not self.timezone.lower() == "utc":
            raise TimeZoneError("Timezones other than UTC are not yet supported")
        return self.ref_date

    @property
    def time_unit(self):
        """Time unit of data"""
        return self.descr_time_unit.split()[0].strip()

    @staticmethod
    def numarr_to_datetime64(basedate, num_arr, mulfac_to_sec):
        """Convert array of numerical timestamps into datetime64 array

        Parameters
        ----------
        basedate : datetime64
            reference date
        num_arr : ndarray
            numerical time stamps relative to basedate
        mulfac_to_sec : float
            multiplicative factor to convert numerical values to unit of
            seconds
        Returns
        -------
        ndarray
            array containing timestamps as datetime64 objects
        """
        totnum = len(num_arr)
        if totnum == 0:
            raise AttributeError("No data available in file")
        return basedate + (num_arr * mulfac_to_sec).astype("timedelta64[s]")

    def all_cols_contain(self, colnums, what):
        """
        Check if all input columns contain input attr `what`

        Parameters
        ----------
        colnums : list
            list of column numbers
        what : str
            name of attribute (e.g. `matrix`, `statistics`,
            `tower_inlet_height`)

        Returns
        -------
        bool
            True if all input columns contain `what` attr., else False.

        """
        return all([what in self.var_defs[col] for col in colnums])

    def assign_flagcols(self):
        _prev = 0
        for idx, item in enumerate(self.var_defs):
            if item.is_flag:
                for _idx in range(_prev, idx):
                    self.var_defs[_idx].flag_col = idx
                _prev = idx + 1

    def init_flags(self, evaluate=True):
        """Decode flag columns and store info in :attr:`flags`"""
        for idx, item in enumerate(self.var_defs):
            if item.is_flag:
                data = self.data[:, idx]
                flag = EbasFlagCol(raw_data=data, interpret_on_init=evaluate)
                self.flag_col_info[idx] = flag

    def compute_time_stamps(self):
        """Compute time stamps from first two data columns"""
        if not self.var_defs[0].unit == self.var_defs[1].unit:
            raise NasaAmesReadError(
                "2nd column is not endtime or does not "
                "have the same unit as 1st column (starttime)"
            )
        offs = self.base_date
        unit = self.time_unit
        if unit not in self.TIMEUNIT2SECFAC:
            raise ValueError(f"Invalid unit for temporal resolution: {unit}")
        mulfac = self.TIMEUNIT2SECFAC[unit]

        start = self.numarr_to_datetime64(offs, self.data[:, 0], mulfac)
        stop = self.numarr_to_datetime64(offs, self.data[:, 1], mulfac)

        dt = stop - start
        # mid timestamps
        self.time_stamps = start + dt * 0.5
        self.start_meas = start
        self.stop_meas = stop

        return (start, stop)

    def get_time_gaps_meas(self, np_freq="s"):
        """Get array with time gaps between individual measurements

        This is computed based on start and stop timestamps, e.g.
        `=dt[0] = start[1] - stop[0]`

        Parameters
        ----------
        np_freq : str
            string specifying output frequency of gap values

        Returns
        -------
        ndarray
            array with time-differences as floating point number in specified
            input resolution
        """
        if self.start_meas is None:
            self.compute_time_stamps()  # assigns start / stop attrs.
        start, stop = self.start_meas, self.stop_meas
        gaps = (start[1:] - stop[:-1]).astype(f"timedelta64[{np_freq}]").astype(float)
        return gaps

    def get_dt_meas(self, np_freq="s"):
        """Get array with time between individual measurements

        This is computed based on start and timestamps, e.g.
        `dt[0] = start[1] - start[0]`

        Parameters
        ----------
        np_freq : str
            string specifying output frequency of gap values

        Returns
        -------
        ndarray
            array with time-differences as floating point number in specified
            input resolution
        """
        if self.start_meas is None:
            self.compute_time_stamps()  # assigns start / stop attrs.
        start = self.start_meas
        dts = (start[1:] - start[:-1]).astype(f"timedelta64[{np_freq}]").astype(float)
        return dts

    def _quality_check(self):
        msgs = ""
        if not len(self.data_header) == len(self.var_defs):
            msgs += (
                "Mismatch between variable definitions in header and "
                "number of data columns in table\n"
            )
        if "timezone" not in self.meta:
            msgs += "Timezone not defined in metadata"
        if msgs:
            raise AttributeError(f"Quality check failed. Messages: {msgs}")

    def read_header(self, nasa_ames_file, quality_check=True):
        self.read_file(nasa_ames_file, only_head=True, quality_check=quality_check)

    def read_file(
        self,
        nasa_ames_file,
        only_head=False,
        replace_invalid_nan=True,
        convert_timestamps=True,
        evaluate_flags=False,
        quality_check=False,
    ):
        """Read NASA Ames file

        Parameters
        ----------
        nasa_ames_file : str
            EBAS NASA Ames file
        only_head : bool
            read only file header
        replace_invalid_nan : bool
            replace all invalid values in the table by NaNs. The invalid values for
            each dependent data column are identified based on the information in
            the file header.
        convert_timestamps : bool
            compute array of numpy datetime64 timestamps from numeric timestamps
            in data
        evaluate_flags : bool
            if True, all data columns get assigned their corresponding flag
            column, the flags in all flag columns are decoded from floating
            point representation to 3 integers, e.g. 0.111222333 -> 111 222 333
            and if input ```replace_invalid_nan==True```, then the invalid
            measurements in each column are replaced with NaN's.
        quality_check : bool
            perform quality check after import (for details see
            :func:`_quality_check`)
        """
        logger.info(f"Reading NASA Ames file:\n{nasa_ames_file}")
        lc = 0  # line counter
        dc = 0  # data block line counter
        mc = 0  # meta block counter
        END_VAR_DEF = np.nan  # will be set (info stored in header)
        IN_DATA = False
        data = []
        self.file = nasa_ames_file
        for line in open(nasa_ames_file):
            if IN_DATA:  # in data block (end of file)
                try:
                    data.append(tuple(float(x.strip()) for x in line.strip().split()))
                    # data.append([float(x.strip()) for x in line.strip().split()])
                except Exception as e:
                    logger.warning(f"EbasNasaAmesFile: Failed to read data row {dc}. Reason: {e}")
                dc += 1
            elif lc < self._NUM_FIXLINES:  # in header section (before column definitions)
                try:
                    val = self._H_FIXLINES_CONV[lc](line)
                    attr = self._H_FIXLINES_YIELD[lc]
                    if isinstance(attr, list):
                        for i, attr_id in enumerate(attr):
                            self[attr_id] = val[i]
                    else:
                        self[attr] = val
                except Exception as e:
                    msg = f"Failed to read header row {lc}.\n{line}\nError msg: {repr(e)}"
                    if lc in self._HEAD_ROWS_MANDATORY:
                        raise NasaAmesReadError(f"Fatal: {msg}")
                    else:
                        logger.warning(msg)
            else:  # behind header section and before data definition (contains column defs and meta info)
                if mc == 0:  # still in column definition
                    END_VAR_DEF = self._NUM_FIXLINES + self.num_cols_dependent - 1
                    NUM_HEAD_LINES = self.num_head_lines
                    try:
                        self.var_defs.append(self._read_vardef_line(line))
                    except Exception as e:
                        logger.warning(repr(e))

                elif lc < END_VAR_DEF:
                    self.var_defs.append(self._read_vardef_line(line))

                elif lc == NUM_HEAD_LINES - 1:
                    IN_DATA = True
                    self._data_header = h = [x.strip() for x in line.split()]
                    # append information of first two columns to variable
                    # definition array.
                    self._var_defs.insert(
                        0, EbasColDef(name=h[0], is_flag=False, is_var=False, unit=self.time_unit)
                    )
                    self._var_defs.insert(
                        1, EbasColDef(name=h[1], is_flag=False, is_var=False, unit=self.time_unit)
                    )
                    if only_head:
                        return
                    logger.debug("REACHED DATA BLOCK")
                elif lc >= END_VAR_DEF + 2:
                    try:
                        name, val = line.split(
                            ":", 1
                        )  # Adding maxpslit=1 incase colon appears in url
                        key = name.strip().lower().replace(" ", "_")
                        self.meta[key] = val.strip()
                    except Exception as e:
                        logger.warning(
                            f"Failed to read line no. {lc}.\n{line}\nError msg: {repr(e)}\n"
                        )
                else:
                    logger.debug(f"Ignoring line no. {lc}: {line}")
                mc += 1
            lc += 1

        data = np.asarray(data)

        data[:, 1:] = data[:, 1:] * np.asarray(self.mul_factors)

        self._data = data
        if replace_invalid_nan:
            dep_dat = data[:, 1:]
            for i, val in enumerate(np.floor(self.vals_invalid)):
                col = dep_dat[:, i]
                cond = np.floor(col) == val
                col[cond] = np.nan
                dep_dat[:, i] = col

            data[:, 1:] = dep_dat
        self._data = data

        if convert_timestamps:
            self.compute_time_stamps()

        self.assign_flagcols()
        self.init_flags(evaluate=evaluate_flags)

        if quality_check:
            self._quality_check()

    def _read_vardef_line(self, line_from_file):
        """Import variable definition line from NASA Ames file"""
        lineX = line_from_file.replace(", ", ",")  # avoid two-char delimiters
        cr = csv.reader(StringIO(lineX), delimiter=",", quotechar='"')
        row = next(cr)
        spl = [x.strip() for x in row]
        spl = [x.replace(",", ", ") for x in spl]  # add space after comma again in quoted fields
        name = spl[0]
        if not len(spl) > 1:
            unit = ""
        else:
            unit = spl[1]
        data = EbasColDef(name=name, is_flag=True, is_var=False, unit=unit)

        if "numflag" not in name:
            data.is_var = True
            data.is_flag = False
            for item in spl[2:]:
                if "=" in item:
                    # e.g. wavelength=550nm
                    sub = item.split("=")
                    if len(sub) == 2:
                        idf, val = (x.strip() for x in sub)
                        data[idf.lower().replace(" ", "_")] = val
                    else:
                        logger.warning(
                            f"Could not interpret part of column "
                            f"definition in EBAS NASA Ames file: {item}"
                        )
                else:  # unit
                    logger.warning(f"Failed to interpret {item}")

        return data

    def _data_short_str(self):
        if len(self.data) == 0:
            s = "No data available"
        else:
            s = str(self.data)
            shape = self.data.shape
            s += f"\nColnum: {shape[1]}"
            s += f"\nTimestamps: {shape[0]}"
        return s

    def print_col_info(self):
        """Print information about individual columns"""
        for idx, coldef in enumerate(self.var_defs):
            print(f"Column {idx}\n{coldef}")

    def __str__(self):
        s = super().__str__()
        s += f"\n\n{str_underline('Data', indent=3)}"
        s += f"\n{self._data_short_str()}"
        return s



import logging

import xarray as xr
from geonum.atmosphere import T0_STD, p0
from pyaerocom.aux_var_helpers import concx_to_vmrx
from pyaerocom.molmasses import get_molmass

logger = logging.getLogger(__name__)


def add_dataarrays(arr0: xr.DataArray, *arrs: xr.DataArray) -> xr.DataArray:
    """
    Add a bunch of :class:`xarray.DataArray` instances

    Parameters
    ----------
    *arr0
        first input array (instance of :class:`xarray.DataArray` with same shape).
    *arrs
        Additional input arrays (instances of :class:`xarray.DataArray` with same shape)

    Returns
    -------
    xarray.DataArray
        Added array
    """
    result = arr0.copy(deep=True)
    for arr in arrs:
        result += arr
    return result


def subtract_dataarrays(arr0: xr.DataArray, *arrs: xr.DataArray) -> xr.DataArray:
    """
    Subtract a bunch of :class:`xarray.DataArray` instances from an array

    Parameters
    ----------
    arr0
        Input array (instance of :class:`xarray.DataArray` with same shape).
    *arrs
        input arrays (instances of :class:`xarray.DataArray` with same shape).
        Subtraction is performed with respect to `arr0`.

    Returns
    -------
    xarray.DataArray
        Diff array (all additional ones are subtracted from `arr0`)
    """
    result = arr0.copy(deep=True)
    for arr in arrs:
        result -= arr
    return result


def calc_concNhno3(conchno3: xr.DataArray) -> xr.DataArray:
    M_N = 14.006
    M_O = 15.999
    M_H = 1.007

    conchno3 = conchno3.copy(deep=True)
    fac = M_N / (M_H + M_N + M_O * 3)
    concNhno3 = conchno3 * fac
    concNhno3.attrs["units"] = "ug N m-3"
    return concNhno3


# ToDo: add docstring
def calc_concNno3pm10(concno3f: xr.DataArray, concno3c: xr.DataArray) -> xr.DataArray:
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + 3 * M_O)
    concno3pm10 = concno3f + concno3c
    concNno3pm10 = concno3pm10 * fac
    concNno3pm10.attrs["var_name"] = "concNno3pm10"
    concNno3pm10.attrs["units"] = "ug N m-3"
    return concNno3pm10


# ToDo: add docstring
def calc_concNno3pm25(
    concno3f: xr.DataArray, concno3c: xr.DataArray, fine_from_coarse_fraction: float = 0.134
) -> xr.DataArray:
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + 3 * M_O)
    concno3pm25 = concno3f + fine_from_coarse_fraction * concno3c
    concNno3pm25 = concno3pm25 * fac
    concNno3pm25.attrs["var_name"] = "concNno3pm25"
    concNno3pm25.attrs["units"] = "ug N m-3"
    return concNno3pm25


# ToDo: add docstring
def calc_concno3pm10(concno3f, concno3c):
    concno3pm10 = concno3f + concno3c

    concno3pm10.attrs["var_name"] = "concno3pm10"
    concno3pm10.attrs["units"] = "ug m-3"

    return concno3pm10


# ToDo: add docstring
def calc_concno3pm25(concno3f, concno3c, fine_from_coarse_fraction=0.134):
    concno3pm25 = concno3f + fine_from_coarse_fraction * concno3c

    concno3pm25.attrs["var_name"] = "concno3pm25"
    concno3pm25.attrs["units"] = "ug m-3"
    return concno3pm25


def calc_conNtno3(
    conchno3: xr.DataArray, concno3f: xr.DataArray, concno3c: xr.DataArray
) -> xr.DataArray:
    concNhno3 = calc_concNhno3(conchno3)
    concNno3pm10 = calc_concNno3pm10(concno3f, concno3c)

    concNtno3 = concNhno3 + concNno3pm10
    concNtno3.attrs["units"] = "ug N m-3"
    return concNtno3


# ToDo: add docstring
def calc_concNnh3(concnh3: xr.DataArray) -> xr.DataArray:
    M_N = 14.006
    M_H = 1.007

    concnh3 = concnh3.copy(deep=True)
    concNnh3 = concnh3 * (M_N / (M_H * 3 + M_N))
    concNnh3.attrs["units"] = "ug N m-3"
    return concNnh3


# ToDo: add docstring
def calc_concNnh4(concnh4: xr.DataArray) -> xr.DataArray:
    M_N = 14.006
    M_H = 1.007

    concnh4 = concnh4.copy(deep=True)
    concNnh4 = concnh4 * (M_N / (M_H * 4 + M_N))
    concNnh4.attrs["units"] = "ug N m-3"
    return concNnh4


# ToDo: add docstring
def calc_concNtnh(concnh3: xr.DataArray, concnh4: xr.DataArray) -> xr.DataArray:
    concNnh3 = calc_concNnh3(concnh3)
    concNnh4 = calc_concNnh4(concnh4)

    concNtnh = concNnh3 + concNnh4
    concNtnh.attrs["units"] = "ug N m-3"
    return concNtnh


# ToDo: add docstring
def update_EC_units(concecpm25: xr.DataArray) -> xr.DataArray:
    concCecpm25 = concecpm25
    concCecpm25.attrs["units"] = "ug C m-3"

    return concCecpm25


def calc_concsspm25(
    concssf: xr.DataArray, concssc: xr.DataArray, coarse_fraction: float = 0.13
) -> xr.DataArray:
    """
    Calculate PM2.5 seasalt

    Parameters
    ----------
    concssf : xr.DataArray
        EMEP output fine seasalt
    concssc : xr.DataArray
        EMEP output coarse seasalt
    coarse_fraction : float
        fraction of coarse supposed to be added to fine output, defaults to
        0.13.

    Returns
    -------
    xr.DataArray
        PM2.5 seasalt including input coarse fraction

    """
    concsspm25 = concssf + coarse_fraction * concssc

    concsspm25.attrs["units"] = "ug m-3"
    return concsspm25


def calc_vmrox(concno2: xr.DataArray, vmro3: xr.DataArray) -> xr.DataArray:
    """
    Calculate OX VMR from NO2 concentration and O3 VMR

    Converts NO2 conc to NO2 VMR assuming US standard atmosphere and adds
    that with VMR O3.

    Parameters
    ----------
    concno2 : xr.DataArray
        mass concentration of NO2
    vmro3 : xr.DataArray
        volume mixing ratio of O3

    Returns
    -------
    xr.DataArray
        volume mixing ratio of OX (O3 + NO2) in units of nmole mole-1

    """
    vmrno2 = concx_to_vmrx(
        data=concno2,
        p_pascal=p0,  # 1013 hPa (US standard atm)
        T_kelvin=T0_STD,  # 15 deg celcius (US standard atm)
        conc_unit=str(concno2.attrs["units"]),
        mmol_var=get_molmass("no2"),  # g/mol NO2
        to_unit="nmol mol-1",
    )

    vmrox = vmrno2 + vmro3
    vmrox.attrs["units"] = "nmol mol-1"
    return vmrox


def calc_vmrox_from_conc(concno2, conco3):
    """
    Calculate OX VMR from NO2 concentration and O3 VMR

    Converts NO2 conc to NO2 VMR assuming US standard atmosphere and adds
    that with VMR O3.

    Parameters
    ----------
    concno2 : xr.DataArray
        mass concentration of NO2
    vmro3 : xr.DataArray
        volume mixing ratio of O3

    Returns
    -------
    xr.DataArray
        volume mixing ratio of OX (O3 + NO2) in units of nmole mole-1

    """
    vmrno2 = concx_to_vmrx(
        data=concno2,
        p_pascal=p0,  # 1013 hPa (US standard atm)
        T_kelvin=T0_STD,  # 15 deg celcius (US standard atm)
        conc_unit=str(concno2.attrs["units"]),
        mmol_var=get_molmass("no2"),  # g/mol NO2
        to_unit="nmol mol-1",
    )

    vmro3 = concx_to_vmrx(
        data=conco3,
        p_pascal=p0,  # 1013 hPa (US standard atm)
        T_kelvin=T0_STD,  # 15 deg celcius (US standard atm)
        conc_unit=str(conco3.attrs["units"]),
        mmol_var=get_molmass("o3"),  # g/mol O3
        to_unit="nmol mol-1",
    )

    vmrox = vmrno2 + vmro3
    vmrox.attrs["units"] = "nmol mol-1"
    return vmrox


def calc_vmrno2(concno2: xr.DataArray) -> xr.DataArray:
    vmrno2 = concx_to_vmrx(
        data=concno2,
        p_pascal=p0,  # 1013 hPa (US standard atm)
        T_kelvin=T0_STD,  # 15 deg celcius (US standard atm)
        conc_unit=str(concno2.attrs["units"]),
        mmol_var=get_molmass("no2"),  # g/mol NO2
        to_unit="nmol mol-1",
    )

    vmrno2.attrs["units"] = "nmol mol-1"
    return vmrno2


def identity(arr):
    return arr


def calc_conNtno3_emep(*arrs):
    if len(arrs) > 1:
        raise ValueError("Should only be given 1 array")

    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + 3 * M_O)

    concNtno3 = arrs[0].copy(deep=True)
    concNtno3 = concNtno3 * fac
    concNtno3.attrs["units"] = "ug N m-3"

    return concNtno3


def calc_conNtnh_emep(*arrs):
    if len(arrs) > 1:
        raise ValueError("Should only be given 1 array")

    M_N = 14.006
    M_H = 1.007

    fac = M_N / (M_H * 4 + M_N)

    conNtnh = arrs[0].copy(deep=True)
    conNtnh = conNtnh * fac
    conNtnh.attrs["units"] = "ug N m-3"

    return conNtnh


def calc_concso4t(concso4, concss):
    factor = 0.077  # 7.7% SO4+ in seasalt
    concso4t = concso4 + factor * concss
    concso4t.attrs["units"] = "ug m-3"

    return concso4t


def calc_concNno(concno):
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + M_O)

    concno = concno.copy(deep=True)
    concNno = concno * fac
    concNno.attrs["units"] = "ug N m-3"

    return concNno


def calc_vmro3(conco3):
    vmro3 = concx_to_vmrx(
        data=conco3,
        p_pascal=p0,  # 1013 hPa (US standard atm)
        T_kelvin=T0_STD,  # 15 deg celcius (US standard atm)
        conc_unit=str(conco3.attrs["units"]),
        mmol_var=get_molmass("o3"),  # g/mol O3
        to_unit="nmol mol-1",
    )

    vmro3.attrs["units"] = "nmol mol-1"
    return vmro3


def calc_concNno2(concno2):
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + M_O * 2)

    concno2 = concno2.copy(deep=True)
    concNno2 = concno2 * fac
    concNno2.attrs["units"] = "ug N m-3"

    return concNno2


def calc_concSso2(concso2):
    M_O = 15.999
    M_S = 32.065
    fac = M_S / (M_S + M_O * 2)

    concso2 = concso2.copy(deep=True)
    concSso2 = concso2 * fac
    concSso2.attrs["units"] = "ug S m-3"

    return concSso2


def calc_concpolyol(concspores):
    # polyol is 4.5% of spores. Spores is in ug/cm3 in Gunnars run, eventhough the unit is marked as ugm-3, so a factor of 1000 is needed for unit change
    factor = 45.0 / 1000.0

    concpolyol = concspores.copy(deep=True) * factor
    concpolyol.attrs["units"] = "ug m-3"
    return concpolyol


def calc_ratpm10pm25(concpm10: xr.DataArray, concpm25: xr.DataArray) -> xr.DataArray:
    """
    Calculate ratio of pm10 and pm25

        Parameters
    ----------
    concpm10 : xr.DataArray
        mass concentration pm10
    concpm25 : xr.DataArray
        mass concentration of pm25

    Returns
    -------
    xr.DataArray
        ratio of concpm10 / concpm25 in units of 1

    """
    try:
        if concpm10.attrs["units"] != concpm25.attrs["units"]:
            logger.warning(
                f"concpm10 unit {concpm10.attrs['units']} not equal to concpm25 unit {concpm25.attrs['units']}!"
            )
    except KeyError:
        pass
    ratpm10pm25 = concpm10 / concpm25
    ratpm10pm25.attrs["units"] = "1"
    return ratpm10pm25


def calc_ratpm25pm10(concpm25: xr.DataArray, concpm10: xr.DataArray) -> xr.DataArray:
    """
    Calculate ratio of pm10 and pm25

        Parameters
    ----------
    concpm10 : xr.DataArray
        mass concentration pm10
    concpm25 : xr.DataArray
        mass concentration of pm25

    Returns
    -------
    xr.DataArray
        ratio of concpm25 / concpm10 in units of 1

    """
    try:
        if concpm10.attrs["units"] != concpm25.attrs["units"]:
            logger.warning(
                f"concpm10 unit {concpm10.attrs['units']} not equal to concpm25 unit {concpm25.attrs['units']}!"
            )
    except KeyError:
        pass
    ratpm25pm10 = concpm25 / concpm10
    ratpm25pm10.attrs["units"] = "1"
    return ratpm25pm10



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyaerocom.exceptions import DataRetrievalError
from pyaerocom.io.read_airnow import ReadAirNow
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData


@pytest.fixture(scope="module")
def reader() -> ReadAirNow:
    return ReadAirNow("AirNowSubset")


def test__FILETYPE(reader: ReadAirNow):
    assert reader._FILETYPE == ".dam"


def test__FILEMASK(reader: ReadAirNow):
    assert reader._FILEMASK == "/monthly/*.dam"


def test__version__(reader: ReadAirNow):
    assert reader.__version__ == "0.12"


def test_FILE_COL_DELIM(reader: ReadAirNow):
    assert reader.FILE_COL_DELIM == "|"


def test_FILE_COL_NAMES(reader: ReadAirNow):
    COLUMNS_IN_DATA_FILES = [
        "date",
        "time",
        "station_id",
        "station_name",
        "time_zone",
        "variable",
        "unit",
        "value",
        "institute",
    ]
    assert sorted(reader.FILE_COL_NAMES) == sorted(COLUMNS_IN_DATA_FILES)


def test_STATION_META_MAP(reader: ReadAirNow):
    # mapping of columns in station metadata file to pyaerocom standard
    METADATA_COLUMNS = dict(
        aqsid="station_id",
        name="station_name",
        lat="latitude",
        lon="longitude",
        elevation="altitude",
        city="city",
        address="address",
        timezone="timezone",
        environment="area_classification",
        populationclass="station_classification",
        modificationdate="modificationdate",
        comment="comment",
    )
    assert reader.STATION_META_MAP == METADATA_COLUMNS


def test_STATION_META_DTYPES(reader: ReadAirNow):
    METADATA_TYPES = dict(
        station_id=str,
        station_name=str,
        latitude=float,
        longitude=float,
        altitude=float,
        city=str,
        address=str,
        timezone=str,
        area_classification=str,
        modificationdate=str,
        station_classification=str,
        comment=str,
    )
    assert reader.STATION_META_DTYPES == METADATA_TYPES


def test_REPLACE_STATNAME(reader: ReadAirNow):
    REPLACE_STATNAME = {"&": "and", "/": " ", ":": " ", ".": " ", "'": ""}
    assert reader.REPLACE_STATNAME == REPLACE_STATNAME


def test_BASEYEAR(reader: ReadAirNow):
    # Years in timestamps in the files are are 2-digit (e.g. 20 for 2020)
    assert reader.BASEYEAR == 2000


def test_DATA_ID(reader: ReadAirNow):
    assert reader.DATA_ID == "AirNow"


def test_SUPPORTED_DATASETS(reader: ReadAirNow):
    assert reader.SUPPORTED_DATASETS == ["AirNow", "AirNowSubset"]


def test_UNIT_MAP(reader: ReadAirNow):
    # units found in file
    UNITS = {
        "C": "celcius",
        "M/S": "m s-1",
        "MILLIBAR": "mbar",
        "MM": "mm",
        "PERCENT": "%",
        "PPB": "ppb",
        "PPM": "ppm",
        "UG/M3": "ug m-3",
        "WATTS/M2": "W m-2",
    }
    assert reader.UNIT_MAP == UNITS


def test_variables(reader: ReadAirNow):
    VARIABLES = dict(
        concbc="BC",
        concpm10="PM10",
        concpm25="PM2.5",
        vmrco="CO",
        vmrnh3="NH3",
        vmrno="NO",
        vmrno2="NO2",
        vmrnox="NOX",
        vmrnoy="NOY",
        vmro3="OZONE",
        vmrso2="SO2",
    )
    assert reader.VAR_MAP == VARIABLES
    assert reader.DEFAULT_VARS == reader.PROVIDES_VARIABLES == list(VARIABLES)


def test_TSTYPE(reader: ReadAirNow):
    assert reader.TS_TYPE == "hourly"


def test_STAT_META_FILENAME(reader: ReadAirNow):
    assert reader.STAT_METADATA_FILENAME == "allStations_20191224.csv"


def test__date_time_str_to_datetime64(reader: ReadAirNow):
    dt = reader._date_time_str_to_datetime64("10/23/20", "13:55")

    assert isinstance(dt, np.datetime64)
    assert dt.dtype == "datetime64[s]"
    assert str(dt) == "2020-10-23T13:55:00"


def test__read_metadata_file(reader: ReadAirNow):
    COLUMNS = [
        "aqsid",
        "name",
        "lat",
        "lon",
        "elevation",
        "city",
        "address",
        "position",
        "timezone",
        "environment",
        "modificationdate",
        "populationclass",
        "comment",
    ]

    cfg = reader._read_metadata_file()
    assert isinstance(cfg, pd.DataFrame)
    assert list(cfg.columns) == COLUMNS
    assert cfg.values.shape == (2588, 13)


def test__correct_station_name(reader: ReadAirNow):
    station_name = "Bla/blub.bla'blub:blaa & blub"
    correct_name = "Bla blub blablub blaa and blub"
    assert reader._correct_station_name(station_name) == correct_name


def test__init_station_metadata(reader: ReadAirNow):
    stations = reader._init_station_metadata()
    assert len(stations) == 2588
    assert all(isinstance(x, dict) for x in stations.values())

    ids = list(stations)
    assert ids[0] == "000010101"
    assert ids[1000] == "160550006"

    station = stations["000010101"]
    assert station["station_name"] == "Duckworth and Ordinance"
    assert station["station_id"] == "000010101"
    assert station["latitude"] == pytest.approx(47.568, rel=1e-3)
    assert station["longitude"] == pytest.approx(-52.702, rel=1e-3)
    assert station["altitude"] == 7

    station = stations["160550006"]
    assert station["station_name"] == "Coeur D Alene - Teom"
    assert station["station_id"] == "160550006"
    assert station["latitude"] == pytest.approx(47.682, rel=1e-3)
    assert station["longitude"] == pytest.approx(-116.766, rel=1e-3)
    assert station["altitude"] == 665


def test_data_dir_exists(reader: ReadAirNow):
    assert Path(reader.data_dir).exists()


def test_get_file_list(reader: ReadAirNow):
    files = reader.get_file_list()
    assert len(files) == 3

    FILE_NAMES = ["202306.dam", "202307.dam", "202308.dam"]
    assert [Path(f).name for f in files] == FILE_NAMES


def test__read_file(reader: ReadAirNow):
    file = reader.get_file_list()[1]
    data = reader._read_file(file)
    assert isinstance(data, pd.DataFrame)
    assert list(data.columns) == reader.FILE_COL_NAMES
    assert data.values.shape == (2000, 9)


# This should test all variables available and reads the first 3 data files
# so for each variable, the StationData objects should contain 3 timestamps
@pytest.mark.parametrize(
    "var_name,statnum,first_dtime,first_vals,unit",
    [
        pytest.param(
            "concbc",
            2,
            ["2023-05-31T16:00:00", "2023-06-30T16:00:00"],
            [0.17, 0.27],
            "ug m-3",
            id="concbc",
        ),
        pytest.param(
            "concpm10",
            98,
            ["2023-05-31T16:00:00", "2023-06-30T16:00:00"],
            [17.0, 5.0],
            "ug m-3",
            id="concpm10",
        ),
        pytest.param(
            "concpm25",
            221,
            ["2023-05-31T20:00:00", "2023-06-30T20:00:00", "2023-07-31T20:00:00"],
            [7.4, 11.2, 3.9],
            "ug m-3",
            id="concpm25",
        ),
        pytest.param(
            "vmrco",
            50,
            ["2023-06-30T20:00:00"],
            [0.2],
            "ppm",
            id="vmrco",
        ),
        pytest.param(
            "vmrno",
            85,
            ["2023-05-31T20:00:00", "2023-06-30T20:00:00", "2023-07-31T20:00:00"],
            [0.0, 0.0, 0.2],
            "ppb",
            id="vmrno",
        ),
        pytest.param(
            "vmrno2",
            96,
            ["2023-05-31T20:00:00", "2023-06-30T20:00:00", "2023-07-31T20:00:00"],
            [0.8, 0.0, 0.2],
            "ppb",
            id="vmrno2",
        ),
        pytest.param(
            "vmrnox",
            48,
            ["2023-06-30T20:00:00"],
            [13.8],
            "ppb",
            id="vmrnox",
        ),
        pytest.param(
            "vmrnoy",
            3,
            ["2023-05-31T17:00:00", "2023-06-30T17:00:00"],
            [2.2, 2.4],
            "ppb",
            id="vmrnoy",
        ),
        pytest.param(
            "vmro3",
            289,
            ["2023-05-31T20:00:00", "2023-06-30T20:00:00", "2023-07-31T20:00:00"],
            [34.0, 27.0, 12.0],
            "ppb",
            id="vmro3",
        ),
        pytest.param(
            "vmrso2",
            57,
            ["2023-05-31T20:00:00", "2023-06-30T20:00:00", "2023-07-31T20:00:00"],
            [0.1, 0.0, 0.0],
            "ppb",
            id="vmrso2",
        ),
    ],
)
def test__read_files_single_var(
    reader: ReadAirNow,
    var_name: str,
    statnum: int,
    first_dtime: list[str],
    first_vals: list[float],
    unit: str,
):
    files = reader.get_file_list()
    data = reader._read_files(files, [var_name])
    assert isinstance(data, list)
    assert len(data) == statnum

    for stat in data:
        assert isinstance(stat, StationData)
        assert var_name in stat

    first_stat = data[0]

    # timeseries metadata
    assert var_name in first_stat["var_info"]
    assert "units" in first_stat["var_info"][var_name]
    assert first_stat["var_info"][var_name]["units"] == unit

    dtimelist = [str(x) for x in first_stat.dtime]
    assert dtimelist == first_dtime

    assert isinstance(first_stat[var_name], np.ndarray)
    assert list(first_stat[var_name]) == first_vals


def test__read_files_single_var_error(reader: ReadAirNow):
    files = reader.get_file_list()
    # NH3 not available in selected 2 test files
    with pytest.raises(DataRetrievalError) as e:
        reader._read_files(files, ["vmrnh3"])
    assert str(e.value) == "None of the input variables could be found in input list"


@pytest.mark.parametrize(
    "vars_to_retrieve, num_meta_blocks,num_stats",
    [
        ("concpm10", 98, 98),
        (
            [
                "concbc",
                "concpm10",
                "concpm25",
                "vmrco",
                "vmrno",
                "vmrno2",
                "vmrnox",
                "vmrnoy",
                "vmro3",
                "vmrso2",
            ],
            949,
            342,
        ),
    ],
)
def test_read(
    reader: ReadAirNow, vars_to_retrieve: str | list[str], num_meta_blocks: int, num_stats: int
):
    data = reader.read(vars_to_retrieve)
    if isinstance(vars_to_retrieve, str):
        vars_to_retrieve = [vars_to_retrieve]
    assert isinstance(data, UngriddedData)
    assert len(data.metadata) == num_meta_blocks
    assert len(data.unique_station_names) == num_stats
    assert sorted(data.contains_vars) == sorted(vars_to_retrieve)



import pytest

from pyaerocom import const
from pyaerocom.io.readungriddedbase import ReadUngriddedBase


class DummyReader(ReadUngriddedBase):
    _FILEMASK = ".txt"
    DATA_ID = "Blaaa"
    __version__ = "0.01"
    PROVIDES_VARIABLES = ["od550aer"]
    REVISION_FILE = const.REVISION_FILE

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id, data_dir)

    @property
    def DEFAULT_VARS(self):
        return self.PROVIDES_VARIABLES

    @property
    def SUPPORTED_DATASETS(self):
        return [self.DATA_ID]

    @property
    def TS_TYPE(self):
        raise NotImplementedError

    def read(self):
        raise NotImplementedError

    def read_file(self):
        raise NotImplementedError


@pytest.fixture(scope="module")
def dummy_reader():
    return DummyReader()


def test___init__template():
    with pytest.raises(TypeError):
        ReadUngriddedBase()


def test___init__dummy():
    dummy = DummyReader()
    assert dummy.data_id == "Blaaa"


@pytest.mark.parametrize(
    "key,val",
    [
        ("_FILEMASK", ".txt"),
        ("__version__", "0.01"),
        ("DATA_ID", "Blaaa"),
        ("SUPPORTED_DATASETS", ["Blaaa"]),
        ("PROVIDES_VARIABLES", ["od550aer"]),
        ("DEFAULT_VARS", ["od550aer"]),
        ("data_id", "Blaaa"),
        ("REVISION_FILE", "Revision.txt"),
        ("AUX_VARS", []),
        ("data_id", "Blaaa"),
    ],
)
def test_DummyReader_attrs(dummy_reader, key, val):
    assert getattr(dummy_reader, key) == val


@pytest.mark.parametrize(
    "key,exception,error",
    [
        pytest.param("TS_TYPE", NotImplementedError, "", id="NotImplementedError"),
        pytest.param(
            "data_dir", ValueError, "Observation network ID Blaaa does not exist", id="ValueError"
        ),
    ],
)
def test_DummyReader_attrs_error(dummy_reader, key, exception, error: str):
    with pytest.raises(exception) as e:
        getattr(dummy_reader, key)
    assert str(e.value) == error



from __future__ import annotations

import os.path

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from pyaerocom import const
from pyaerocom.io.gaw.reader import ReadGAW
from tests.conftest import TEST_RTOL


def skip_flag():
    # for some reason const.OBSLOCS_UNGRIDDED[const.DMS_AMS_CVO_NAME] is not defined on CI
    if const.DMS_AMS_CVO_NAME not in const.OBSLOCS_UNGRIDDED:
        return True
    elif os.path.exists(const.OBSLOCS_UNGRIDDED[const.DMS_AMS_CVO_NAME]):
        return False
    else:
        return True


skip = pytest.mark.skipif(
    skip_flag(),
    reason="GAW path not found (are we on CI?)",
    allow_module_level=True,
)


@pytest.fixture(scope="module")
def gaw_path() -> str:
    return const.OBSLOCS_UNGRIDDED[const.DMS_AMS_CVO_NAME]


@pytest.fixture(scope="module")
def _make_data(gaw_path: str) -> ReadGAW:
    r = ReadGAW(data_dir=gaw_path)
    return r.read("vmrdms")


@pytest.fixture(scope="module")
def data_vmrdms_ams_cvo(_make_data, gaw_path):
    return _make_data


@skip
def test_ungriddeddata_ams_cvo(data_vmrdms_ams_cvo):
    data = data_vmrdms_ams_cvo
    assert data.shape == (15616, 12)
    assert len(data.metadata) == 6

    unique_coords = []
    unique_coords.extend(np.unique(data.latitude))
    unique_coords.extend(np.unique(data.longitude))
    unique_coords.extend(np.unique(data.altitude))
    assert len(unique_coords) == 6
    assert_allclose(unique_coords, [-37.8, 16.848, -24.871, 77.53, 10.0, 65.0], rtol=TEST_RTOL)

    vals = data._data[:, data.index["data"]]
    assert_allclose(
        [np.nanmean(vals), np.nanstd(vals), np.nanmax(vals), np.nanmin(vals)],
        [1.794279e14, 2.352963e14, 2.807600e15, -7.900000e12],
        rtol=TEST_RTOL,
    )


@skip
def test_vmrdms_ams(data_vmrdms_ams_cvo):
    stat = data_vmrdms_ams_cvo.to_station_data(meta_idx=0)

    keys = list(stat)
    assert "vmrdms" in keys
    assert "var_info" in keys

    assert stat.dtime.min() == np.datetime64("1987-03-01T00:00:00.000000000")
    assert stat.dtime.max() == np.datetime64("2008-12-31T00:00:00.000000000")

    assert stat["instrument_name"] == "unknown"
    assert stat["ts_type"] == "daily"
    assert stat["filename"] == "ams137s00.lsce.as.fl.dimethylsulfide.nl.da.dat"

    vmrdms = stat["vmrdms"]
    assert_allclose(
        [vmrdms.mean(), vmrdms.std(), vmrdms.max(), vmrdms.min()],
        [1.856801e14, 2.371294e14, 2.807600e15, 5.100000e12],
        rtol=TEST_RTOL,
    )


@skip
def test_vmrdms_ams_subset(data_vmrdms_ams_cvo):
    stat = data_vmrdms_ams_cvo.to_station_data(meta_idx=0, start=2000, stop=2008, freq="monthly")

    assert_array_equal(
        [str(stat.dtime.min()), str(stat.dtime.max())],
        ["2000-01-15T00:00:00.000000000", "2007-12-15T00:00:00.000000000"],
    )
    assert stat.ts_type == "monthly"
    assert stat.ts_type_src == "daily"



from __future__ import annotations

from pathlib import Path

from pyaerocom.config import ALL_REGION_NAME

from .common import add_dummy_model_data

YEAR = "2007"


def fake_model_data(tmp_path: str | Path) -> dict:
    add_dummy_model_data(
        "vmrno2",
        "nmole mole-1",
        "monthly",
        "Surface",
        year=YEAR,
        lat_range=(-90, 90),
        lon_range=(-180, 180),
        tmp_path=tmp_path,
    )
    model_data_dir = add_dummy_model_data(
        "vmro3",
        "nmole mole-1",
        "monthly",
        "Surface",
        year=YEAR,
        lat_range=(-90, 90),
        lon_range=(-180, 180),
        tmp_path=tmp_path,
    )
    return dict(
        DUMMY=dict(
            model_id="DUMMY-MODEL",
            model_data_dir=model_data_dir,
        )
    )


OBS_GROUNDBASED = {"EBAS": dict(obs_id="EBASSubset", obs_vars=("vmro3",), obs_vert_type="Surface")}

CFG = dict(
    model_cfg=dict(),  # fake_model_data("PATH_TO_MODEL_DATA"),
    obs_cfg=OBS_GROUNDBASED,
    json_basedir="PATH_TO_AEROVAL_OUT/data",
    coldata_basedir="PATH_TO_AEROVAL_OUT/coldata",
    # if True, existing colocated data files will be deleted
    reanalyse_existing=True,
    raise_exceptions=True,
    only_json=False,
    add_model_maps=False,
    only_model_maps=False,
    clear_existing_json=False,
    # Regional filter for analysis
    filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="monthly",
    map_zoom="World",
    freqs=["monthly"],
    periods=[YEAR],
    main_freq="monthly",
    harmonise_units=True,
    proj_id="test",
    exp_id="exp3",
    exp_name="AeroVal test experiment 3",
    exp_descr=("Test setup for more complex evaluation configurations"),
    exp_pi="Jonas Gliss",
    public=True,
    # directory where colocated data files are supposed to be stored
    weighted_stats=True,
)



from pathlib import Path
from textwrap import dedent

from pydantic import ValidationError
from pytest import mark, param, raises

from pyaerocom.aeroval.aux_io_helpers import ReadAuxHandler, check_aux_info


def test_ReadAuxHandler(tmp_path: Path):
    path = tmp_path / "dummy.py"
    assert not path.exists()

    dummy = """
    def success() -> bool:
        return True
    
    FUNS = [success]
    """
    path.write_text(dedent(dummy))

    aux = ReadAuxHandler(str(path))
    success = aux.import_module().success
    assert success()

    funcs = aux.import_all()
    assert funcs[0] is success


def test_ReadAuxHandler_empty(tmp_path: Path):
    path = tmp_path / "empty.py"
    assert not path.exists()

    path.write_text("")

    with raises(Exception) as e:
        ReadAuxHandler(str(path)).import_all()

    assert str(e.value) == f"module '{path.stem}' has no attribute 'FUNS'"


@mark.parametrize(
    "fun,vars_required,funcs",
    [
        param(lambda: True, ["first"], {}, id="func as first arg"),
        param("func", ["last"], {"func": lambda: True}, id="func as last arg"),
    ],
)
def test_check_aux_info(fun, vars_required: list[str], funcs: dict):
    aux = check_aux_info(fun, vars_required, funcs)
    assert aux["fun"] is funcs.get(fun, fun)
    assert aux["vars_required"] == vars_required


@mark.parametrize(
    "fun,vars_required,funcs,error,",
    [
        param(
            None,
            [],
            {},
            "2 validation errors for _AuxReadSpec",
            id="no func",
        ),
        param(
            None,
            [42],
            {},
            "3 validation errors for _AuxReadSpec",
            id="bad type vars_required",
        ),
    ],
)
def test_check_aux_info_error(fun, vars_required: list[str], funcs: dict, error: str):
    with raises(ValidationError) as e:
        check_aux_info(fun, vars_required, funcs)

    assert error in str(e.value)



from collections.abc import Callable

import numpy as np

# Type definition for a callable which filters (ie. excludes) data before calculating stats.
DataFilter = Callable[[np.ndarray, np.ndarray, np.ndarray | None], np.ndarray]

# Type definition for a callable which calculates a statistic.
StatisticsCalculator = Callable[[np.ndarray, np.ndarray, np.ndarray | None], np.float64]

# Type definition for a stats dictionary.
StatsDict = dict[str, np.float64]

# Type definition for a callable which filters out statistics from the resulting stats dictionary.
StatisticsFilter = Callable[[StatsDict], StatsDict]



from __future__ import annotations

import pytest
from pandas import Timestamp

from pyaerocom.aeroval.helpers import (
    _check_statistics_periods,
    _get_min_max_year_periods,
    _period_str_to_timeslice,
    check_if_year,
    make_dummy_model,
)
from pyaerocom.aeroval import EvalSetup


@pytest.mark.parametrize(
    "period,result",
    [
        (["2010-2010", "2005"], ["2010-2010", "2005"]),
        (["20100221-20100912", "20230209"], ["2010/02/21-2010/09/12", "2023/02/09"]),
    ],
)
def test__check_statistics_periods(period: list[str], result: list[str]):
    assert _check_statistics_periods(period) == result


@pytest.mark.parametrize(
    "periods,error",
    [
        pytest.param(
            42,
            "statistics_periods needs to be a list",
            id="not list",
        ),
        pytest.param(
            [42],
            "All periods need to be strings",
            id="not list[str]",
        ),
        pytest.param(
            ["2010-2010-20"],
            "Invalid value for period (2010-2010-20), can be either single years/dates or range of years/dates (e.g. 2000-2010).",
            id="wrong period",
        ),
    ],
)
def test__check_statistics_periods_error(periods: list[str], error: str):
    with pytest.raises(ValueError) as e:
        _check_statistics_periods(periods)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "period,result",
    [
        ("2005", slice("2005", "2005")),
        ("2005-2019", slice("2005", "2019")),
    ],
)
def test__period_str_to_timeslice(period: str, result: slice):
    assert _period_str_to_timeslice(period) == result


def test__period_str_to_timeslice_error():
    with pytest.raises(ValueError) as e:
        _period_str_to_timeslice("2005-2019-2000")
    assert str(e.value) == "2005-2019-2000"


@pytest.mark.parametrize(
    "periods,result",
    [
        (
            ["2005", "2000"],
            (Timestamp("2000-03-16 00:00:00"), Timestamp("2005-03-23 00:00:00")),
        ),
        (
            ["20240316-20240323"],
            (Timestamp("2021-03-16 00:00:00"), Timestamp("2024-03-23 00:00:00")),
        ),
        (
            ["20240316-20240316"],
            (Timestamp("2021-03-16 00:00:00"), Timestamp("2024-03-16 00:00:00")),
        ),
        (
            ["2021-2024", "2021", "2022", "2023", "2024"],
            (Timestamp("2021-01-01 00:00:00"), Timestamp("2024-01-01 00:00:00")),
        ),
        (
            ["2005", "2000", "1999-2021"],
            (Timestamp("1999-01-01 00:00:00"), Timestamp("2021-01-01 00:00:00")),
        ),
    ],
)
def test__get_min_max_year_periods(periods: list[str], result: tuple[int, int]):
    _get_min_max_year_periods(periods) == result


@pytest.mark.parametrize(
    "periods,result",
    [
        (["2005", "2000"], True),
        (["20240316-20240323"], False),
        (["2021-2024", "2021"], True),
    ],
)
def test__check_if_year(periods: list[str], result: bool):
    assert check_if_year(periods) == result


@pytest.mark.parametrize(
    "periods,error",
    [
        (["2005-20000112"], "2005 not on the same format as 20000112"),
        (
            ["1999-2000-2009"],
            "Invalid value for period (1999-2000-2009), can be either single years/dates or range of years/dates (e.g. 2000-2010).",
        ),
        (
            ["2021-2024", "2021", "20231201-20240312"],
            "Found mix of years and dates in ['2021-2024', '2021', '20231201-20240312']",
        ),
    ],
)
def test__check_if_year_error(periods: list[str], error: str):
    with pytest.raises(ValueError) as e:
        check_if_year(periods)
    assert str(e.value) == error


def test__get_min_max_year_periods_error():
    with pytest.raises(ValueError) as e:
        _get_min_max_year_periods(["2005-2004-23", "2000", "1999-2021"])
    assert str(e.value) == "2005-2004-23"


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_make_dummy_model(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    assert cfg.obs_cfg.get_entry("AERONET-Sun")
    model_id = make_dummy_model(["AERONET-Sun"], cfg)
    assert model_id == "dummy_model"



import json
from pathlib import Path

import numpy as np
import pytest

from pyaerocom.aeroval.json_utils import (
    read_json,
    round_floats,
    set_float_serialization_precision,
    write_json,
)


@pytest.fixture()
def json_path(tmp_path: Path) -> Path:
    path = tmp_path / "file.json"
    assert not path.exists()
    return path


@pytest.mark.parametrize(
    "raw,precision,rounded",
    [
        pytest.param(
            1.12344567890,
            5,
            1.12345,
            id="single float",
        ),
        pytest.param(
            [np.float64(2.3456789), np.float32(3.456789012)],
            3,
            [2.346, pytest.approx(3.457, 1e-3)],
            id="np.float list",
        ),
        pytest.param(
            (np.float128(4.567890123), np.float64(5.6789012345)),
            5,
            [pytest.approx(4.56789, 1e-5), 5.67890],
            id="np.float tuple",
        ),
        pytest.param(
            dict(bla=np.float128(0.1234455667), blubb=1, ha="test"),
            5,
            dict(bla=pytest.approx(0.12345, 1e-5), blubb=1, ha="test"),
            id="mixed dict",
        ),
    ],
)
def test_round_floats(raw, precision: int, rounded):
    set_float_serialization_precision(precision)
    _rounded = round_floats(raw)
    if isinstance(raw, list | tuple):
        assert type(_rounded) is list
    if isinstance(raw, dict):
        assert type(_rounded) is dict
    assert _rounded == rounded


def test_read_json(json_path: Path):
    data = {"bla": 42}
    json_path.write_text(json.dumps(data))
    assert json_path.exists()
    assert read_json(json_path) == data


@pytest.mark.parametrize("data", [{"bla": 42}, {"bla": 42, "blub": np.nan}])
@pytest.mark.parametrize("kwargs", [dict(), dict(ignore_nan=True, indent=5)])
def test_write_json(json_path: Path, data: dict, kwargs: dict):
    write_json(data, json_path, **kwargs)
    assert json_path.exists()


def test_write_json_error(json_path: Path):
    with pytest.raises(TypeError) as e:
        write_json({"bla": 42}, json_path, bla=42)
    assert str(e.value).endswith("unexpected keyword argument 'bla'")



import json
from enum import Enum
from pathlib import Path
from shutil import copy2
from typing import Optional

import typer
from pyaerocom import __package__, __version__, change_verbosity, const, download_minimal_dataset
from pyaerocom.aeroval import EvalSetup, ExperimentProcessor
from pyaerocom.io.cachehandler_ungridded import list_cache_files
from pyaerocom.io.utils import browse_database

main = typer.Typer()


def version_callback(value: bool):
    if not value:
        return

    typer.echo(f"🦄 {__package__} {__version__}")
    raise typer.Exit()


@main.callback()
def callback(
    version: Optional[bool] = typer.Option(None, "--version", "-V", callback=version_callback),
):
    """🦄 Pyaerocom Command Line Interface"""


@main.command()
def browse(database: str = typer.Argument(..., help="Provide database name.")):
    """Browse database (e.g., browse <DATABASE>)"""
    print(f"Searching database for matches of {database}")
    print(browse_database(database))


@main.command()
def clearcache():
    """Delete cached data objects"""

    delete = typer.confirm("Are you sure you want to delete all cached data objects?")
    if delete:
        print("Okay then.... Here we go!")
        for path in list_cache_files():
            path.unlink()
    else:
        print("Wise decision, pyaerocom will handle it for you automatically anyways 😜")


@main.command()
def listcache():
    """List cached data objects"""
    for path in list_cache_files():
        typer.echo(str(path))


@main.command()
def ppiaccess():
    """Check if MET Norway's PPI can be accessed"""
    print("True") if const.has_access_lustre else print("False")


class Verbosity(str, Enum):
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"

    def __str__(self) -> str:
        return self.value


@main.command()
def aeroval(
    config: Path = typer.Argument(
        ..., help="experiment configuration (JSON)", exists=True, readable=True
    ),
    reuse_coldata: bool = typer.Option(
        True, "--reuse-coldata/--clean-coldata", help="reuse/clean colocated data before running"
    ),
    verbosity: Verbosity = typer.Option(Verbosity.ERROR, help="console logger level"),
):
    """Run an AeroVal experiment as described in a json config file"""

    if config.suffix != ".json":  # pragma:no cover
        typer.echo(f"{config.suffix=}  != '.json'")
        raise typer.Abort()

    change_verbosity(verbosity)
    CFG = json.loads(config.read_text())
    stp = EvalSetup(**CFG)
    proc = ExperimentProcessor(stp)
    proc.exp_output.delete_experiment_data(also_coldata=not reuse_coldata)
    proc.run()


@main.command()
def getsampledata(
    extract_dir: Path = typer.Option(
        default="./data", help="Folder where data should be extracted", writable=True
    ),
    verbosity: Verbosity = typer.Option(Verbosity.ERROR, help="console logger level"),
):
    """Downloads a minimal sample dataset."""
    download_minimal_dataset(extract_dir_override=extract_dir)


@main.command()
def init():
    """init ~/MyPyaerocom directory and copy the default paths.ini there"""

    mypyaerocom = Path.home() / "MyPyaerocom"
    mypyaerocom.mkdir(exist_ok=True)

    # copy default paths.ini if it doesn't exit
    ini_in_path = Path(__file__).parents[1].joinpath("data/paths.ini")
    ini_out_path = mypyaerocom / "paths.ini"
    if not ini_in_path.exists():
        print(f"Error: {ini_in_path} does not exist. Something is wrong with your pyaerocom installation!")
    if ini_out_path.exists():
        print(f"Error: {ini_out_path} already exists. Please delete it by hand if you really want to overwrite that with the default from pyaerocom.")
    else:
        try:
            copy2(ini_in_path, mypyaerocom)
            print(f"{ini_out_path} successfully created")
        except Exception as e:
            print(f"Error: file {ini_out_path} could not created. The error message was "
                  f"{e}.")


if __name__ == "__main__":
    main()



import cartopy.crs as ccrs
import matplotlib
import matplotlib.pyplot as plt
from cartopy.mpl.geoaxes import GeoAxes
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap, to_hex
from seaborn import color_palette
import io
import xarray

try:
    from geojsoncontour import contourf_to_geojson
except ModuleNotFoundError:
    contourf_to_geojson = None

from pyaerocom.aeroval.coldatatojson_helpers import _get_jsdate
from pyaerocom.helpers import make_datetime_index
from pyaerocom.tstype import TsType


# names of modelmaps type options
CONTOUR = "contour"
OVERLAY = "overlay"


def _jsdate_list(data):
    tst = TsType(data.ts_type)
    idx = make_datetime_index(data.start, data.stop, tst.to_pandas_freq())
    return _get_jsdate(idx.values).tolist()


def calc_contour_json(data, cmap, cmap_bins):
    """
    Convert gridded data into contours for json output

    Parameters
    ----------
    data : GriddedData
        input data
    cmap : str
        colormap of output
    cmap_bins : list
        list containing the bins to which the values are mapped.

    Returns
    -------
    dict
        dictionary containing contour data

    """
    if contourf_to_geojson is None:
        raise ModuleNotFoundError(
            "Map processing for aeroval interface requires "
            "the geojsoncontour package which is not part of the "
            "standard conda installation of pyaerocom."
        )

    plt.close("all")
    matplotlib.use("Agg")
    GeoAxes._pcolormesh_patched = Axes.pcolormesh

    cm = ListedColormap(color_palette(cmap, len(cmap_bins) - 1))

    proj = ccrs.PlateCarree()
    ax = plt.axes(projection=proj)

    try:
        data.check_dimcoords_tseries()
    except Exception:
        data.reorder_dimensions_tseries()

    nparr = data.cube.data
    lats = data.latitude.points
    lons = data.longitude.points

    geojson = {}
    tst = _jsdate_list(data)
    for i, date in enumerate(tst):
        datamon = nparr[i]
        contour = ax.contourf(
            lons,
            lats,
            datamon,
            transform=proj,
            colors=cm.colors,
            levels=cmap_bins,
            extend="max",
        )

        result = contourf_to_geojson(contourf=contour)

        geojson[str(date)] = eval(result)

    ax.figure.colorbar(contour, ax=ax)
    colors_hex = [to_hex(val) for val in cm.colors]

    geojson["legend"] = {
        "colors": colors_hex,
        "levels": list(cmap_bins),
        "var_name": data.var_name,
        "units": str(data.units),
    }

    plt.close("all")
    return geojson


def plot_overlay_pixel_maps(
    data: xarray.DataArray, cmap: str, cmap_bins: list[float], format: str
):  # pragma: no cover
    plt.close("all")
    matplotlib.use("Agg")

    fig, axis = plt.subplots(
        1,
        1,
        subplot_kw=dict(projection=ccrs.Mercator()),
        figsize=(8, 8),
    )

    data.plot(
        ax=axis,
        transform=ccrs.PlateCarree(),
        add_colorbar=False,
        add_labels=False,
        vmin=cmap_bins[0],
        vmax=cmap_bins[-1],
        cmap=cmap,
    )

    with io.BytesIO() as buffer:  # use buffer memory
        plt.savefig(
            buffer,
            bbox_inches="tight",
            transparent=True,
            format=format,
        )
        buffer.seek(0)
        image = buffer.getvalue()

    plt.close("all")

    return image



import string
from pathlib import Path

import numpy as np
import pytest

from pyaerocom import UngriddedData, ungriddeddata
from pyaerocom.exceptions import DataCoverageError, VariableDefinitionError
from tests.fixtures.stations import FAKE_STATION_DATA


@pytest.fixture(scope="module")
def ungridded_empty():
    return UngriddedData()


def test_init_shape(ungridded_empty):
    assert ungridded_empty.shape == (10000000, 12)


def test_init_add_cols():
    d1 = UngriddedData(num_points=2, add_cols=["bla", "blub"])
    assert d1.shape == (2, 14)


def test_add_chunk(ungridded_empty):
    ungridded_empty.add_chunk(111002)
    assert ungridded_empty.shape == (20000000, 12)


def test_coordinate_access():
    d = UngriddedData()

    stat_names = list(string.ascii_lowercase)
    lons = np.arange(len(stat_names))
    lats = np.arange(len(stat_names)) - 90
    alts = np.arange(len(stat_names)) * 13

    for i, n in enumerate(stat_names):
        d.metadata[i] = dict(
            data_id="testcase",
            station_name=n,
            latitude=lats[i],
            longitude=lons[i],
            altitude=alts[i],
        )

    assert d.station_name == stat_names
    assert all(d.latitude == lats)
    assert all(d.longitude == lons)
    assert all(d.altitude == alts)

    with pytest.raises(DataCoverageError):
        d.to_station_data("a")

    c = d.station_coordinates
    assert c["station_name"] == stat_names
    assert all(c["latitude"] == lats)
    assert all(c["longitude"] == lons)
    assert all(c["altitude"] == alts)


def test_check_index_aeronet_subset(aeronetsunv3lev2_subset):
    aeronetsunv3lev2_subset._check_index()


@pytest.mark.dependency
def test_check_set_country(aeronetsunv3lev2_subset):
    idx, countries = aeronetsunv3lev2_subset.check_set_country()
    assert len(idx) == len(aeronetsunv3lev2_subset.metadata)
    assert len(countries) == len(idx)
    assert countries == [
        "Italy",
        "Japan",
        "Mali",
        "Brazil",
        "American Samoa",
        "Fr. S. and Antarctic Lands",
        "Republic of Korea",
        "France",
        "Portugal",
        "France",
        "Barbados",
        "United Kingdom",
        "Bolivia",
        "United States",
        "Fr. Polynesia",
        "China",
        "Taiwan",
        "Algeria",
        "Netherlands",
        "Greece",
        "Belgium",
        "Argentina",
    ]
    idx, countries = aeronetsunv3lev2_subset.check_set_country()
    assert idx == []
    assert countries == []


@pytest.mark.dependency(depends=["test_check_set_country"])
def test_countries_available(aeronetsunv3lev2_subset):
    assert aeronetsunv3lev2_subset.countries_available == [
        "Algeria",
        "American Samoa",
        "Argentina",
        "Barbados",
        "Belgium",
        "Bolivia",
        "Brazil",
        "China",
        "Fr. Polynesia",
        "Fr. S. and Antarctic Lands",
        "France",
        "Greece",
        "Italy",
        "Japan",
        "Mali",
        "Netherlands",
        "Portugal",
        "Republic of Korea",
        "Taiwan",
        "United Kingdom",
        "United States",
    ]


@pytest.mark.dependency(depends=["test_check_set_country"])
@pytest.mark.parametrize(
    "region_id,check_mask,check_country_meta,num_meta",
    [("Italy", True, True, 1), ("EUROPE", True, True, 7), ("OCN", True, True, 8)],
)
def test_filter_region(
    aeronetsunv3lev2_subset, region_id, check_mask, check_country_meta, num_meta
):
    subset = aeronetsunv3lev2_subset.filter_region(
        region_id, check_mask=check_mask, check_country_meta=check_country_meta
    )

    assert len(subset.metadata) == num_meta


# sites in aeronet data

ALL_SITES = [
    "AAOT",
    "ARIAKE_TOWER",
    "Agoufou",
    "Alta_Floresta",
    "American_Samoa",
    "Amsterdam_Island",
    "Anmyon",
    "Avignon",
    "Azores",
    "BORDEAUX",
    "Barbados",
    "Blyth_NOAH",
    "La_Paz",
    "Mauna_Loa",
    "Tahiti",
    "Taihu",
    "Taipei_CWB",
    "Tamanrasset_INM",
    "The_Hague",
    "Thessaloniki",
    "Thornton_C-power",
    "Trelew",
]


@pytest.mark.parametrize(
    "args,sitenames",
    [
        ({"station_name": ["Tr*", "Mauna*"]}, ["Trelew", "Mauna_Loa"]),
        (
            {"station_name": ["Tr*", "Mauna*"], "negate": "station_name"},
            [x for x in ALL_SITES if x not in ["Trelew", "Mauna_Loa"]],
        ),
        (
            {"altitude": [0, 1000], "negate": "altitude"},
            ["La_Paz", "Mauna_Loa", "Tamanrasset_INM"],
        ),
        ({"station_name": "Tr*"}, ["Trelew"]),
        (
            {"station_name": "Tr*", "negate": "station_name"},
            [x for x in ALL_SITES if not x == "Trelew"],
        ),
    ],
)
def test_filter_by_meta(aeronetsunv3lev2_subset, args, sitenames):
    data = aeronetsunv3lev2_subset
    subset = data.filter_by_meta(**args)
    sites = [x["station_name"] for x in subset.metadata.values()]
    stats = sorted(list(dict.fromkeys(sites)))
    assert sorted(sitenames) == stats


def test_cache_reload(aeronetsunv3lev2_subset: UngriddedData, tmp_path: Path):
    path = tmp_path / "ungridded_aeronet_subset.pkl"
    file = aeronetsunv3lev2_subset.save_as(file_name=path.name, save_dir=path.parent)
    assert Path(file) == path
    assert path.exists()
    data = UngriddedData.from_cache(data_dir=path.parent, file_name=path.name)
    assert data.shape == aeronetsunv3lev2_subset.shape


def test_check_unit(data_scat_jungfraujoch):
    data_scat_jungfraujoch.check_unit("sc550aer", unit="1/Mm")
    from pyaerocom.exceptions import MetaDataError

    with pytest.raises(MetaDataError):
        data_scat_jungfraujoch.check_unit("sc550aer", unit="m-1")


@pytest.mark.filterwarnings("ignore:invalid value encountered in .*divide:RuntimeWarning")
def test_check_convert_var_units(data_scat_jungfraujoch):
    out = data_scat_jungfraujoch.check_convert_var_units("sc550aer", "m-1", inplace=False)

    fac = 1e-6
    data_idx = out._DATAINDEX
    for i, meta in out.metadata.items():
        if "sc550aer" in meta["var_info"]:
            assert meta["var_info"]["sc550aer"]["units"] == "m-1"
            idx = out.meta_idx[i]["sc550aer"]

            data0 = data_scat_jungfraujoch._data[idx, data_idx]
            data1 = out._data[idx, data_idx]

            ratio = np.divide(data1, data0)  # [~nans]
            ratio = ratio[~np.isnan(ratio)]
            assert ratio.mean() == pytest.approx(fac)
            assert ratio.std() == pytest.approx(0)


def test_from_single_station_data():
    stat = FAKE_STATION_DATA["station_data1"]
    d = ungriddeddata.UngriddedData.from_station_data(stat)
    data0 = stat.ec550aer
    data1 = d.all_datapoints_var("ec550aer")
    assert data0 == pytest.approx(data1, abs=1e-20)


def test_last_meta_idx(aeronetsunv3lev2_subset: UngriddedData):
    assert isinstance(aeronetsunv3lev2_subset.last_meta_idx, np.ndarray | np.generic)


def test_has_flag_data(aeronetsunv3lev2_subset: UngriddedData):
    assert isinstance(aeronetsunv3lev2_subset.has_flag_data, np.bool_ | bool)


def test_is_filtered(aeronetsunv3lev2_subset: UngriddedData):
    assert isinstance(aeronetsunv3lev2_subset.is_filtered, np.bool_ | bool)


def test_available_meta_keys(aeronetsunv3lev2_subset: UngriddedData):
    assert isinstance(aeronetsunv3lev2_subset.available_meta_keys, list)
    assert all(isinstance(key, str) for key in aeronetsunv3lev2_subset.available_meta_keys)


def test_nonunique_station_names(aeronetsunv3lev2_subset: UngriddedData):
    assert isinstance(aeronetsunv3lev2_subset.nonunique_station_names, list)


def test_set_flags_nan_error(aeronetsunv3lev2_subset: UngriddedData):
    data = aeronetsunv3lev2_subset.copy()
    with pytest.raises(AttributeError):
        data = data.data.set_flags_nan(inplace=True)


def test_remove_outliers(aeronetsunv3lev2_subset: UngriddedData):
    data = aeronetsunv3lev2_subset.copy()
    assert not data.filter_hist
    new = data.remove_outliers(var_name="od550aer", low=0, high=0)
    assert new.filter_hist


def test_extract_var(aeronetsunv3lev2_subset: UngriddedData):
    data = aeronetsunv3lev2_subset.copy()
    od = data.extract_var("od550aer")
    assert not data.is_filtered
    assert od.is_filtered
    assert od.shape[0] < data.shape[0]


def test_extract_var_error(aeronetsunv3lev2_subset: UngriddedData):
    data = aeronetsunv3lev2_subset.copy()
    with pytest.raises(VariableDefinitionError):
        data.extract_var("nope")


def test_find_common_stations(aeronetsunv3lev2_subset: UngriddedData):
    data1 = aeronetsunv3lev2_subset.copy()
    data2 = aeronetsunv3lev2_subset.copy()
    station_map = data1.find_common_stations(other=data2)
    assert station_map == {key: key for key in station_map}



# ToDo: merge with test_coldatatojson_helpers.py
from __future__ import annotations

import numpy as np
import pytest
import xarray

from pyaerocom import ColocatedData, TsType
from pyaerocom.aeroval.coldatatojson_helpers import (
    _create_diurnal_weekly_data_object,
    _get_jsdate,
    _get_period_keys,
    _init_data_default_frequencies,
    _init_meta_glob,
    _make_trends,
    _map_indices,
    _process_statistics_timeseries,
    _remove_less_covered,
)
from pyaerocom.aeroval.exceptions import TrendsError
from pyaerocom.exceptions import TemporalResolutionError, UnknownRegion
from tests.fixtures.collocated_data import COLDATA


def get_profile_filename():
    json = get_profile_filename("reg1", "obs1", "var1")
    assert json == "reg1_obs1_var1.json"


@pytest.mark.parametrize(
    "to_ts_types",
    [
        ["daily", "monthly"],
        ["3yearly"],
    ],
)
@pytest.mark.parametrize("coldataset", ["tm5_aeronet"])
def test__init_data_default_frequencies(coldata: ColocatedData, to_ts_types: str):
    result = _init_data_default_frequencies(coldata, to_ts_types)
    assert len(result) == len(to_ts_types)

    tst = TsType(coldata.ts_type)
    for freq, val in result.items():
        if TsType(freq) > tst:
            assert val is None
        else:
            assert isinstance(val, ColocatedData)
            assert val.ts_type == freq


@pytest.fixture(scope="module")
def example_coldata():
    """dictionary of colocated data statistics"""
    return _init_data_default_frequencies(COLDATA["tm5_aeronet"](), ["daily", "monthly", "yearly"])


def test_get_jsdate(example_coldata):
    time = _get_jsdate(example_coldata["monthly"].data.time.values)
    assert len(time) == 12


@pytest.mark.parametrize(
    "freq,region_ids,data_freq,nmb_avg",
    [
        ("yearly", {"EUROPE": "Europe"}, "monthly", 0.168),
        ("yearly", {"EUROPE": "Europe"}, None, 0.122),
        ("monthly", {"EUROPE": "Europe"}, None, 0.181),
        ("monthly", {"EUR": "Europe(HTAP)"}, None, 0.181),
        ("monthly", {"SEA": "SE Asia(HTAP)"}, None, np.nan),
        ("monthly", {"SEA": "SE Asia(HTAP)", "EUR": "Europe(HTAP)"}, None, 0.181),
        ("monthly", {}, None, np.nan),
    ],
)
@pytest.mark.filterwarnings("ignore:Mean of empty slice:RuntimeWarning")
def test__process_statistics_timeseries(
    example_coldata, freq: str, region_ids: dict[str, str], data_freq: str, nmb_avg
):
    result = _process_statistics_timeseries(
        data=example_coldata,
        freq=freq,
        region_ids=region_ids,
        use_weights=False,
        drop_stats=(),
        use_country=False,
        data_freq=data_freq,
    )
    assert len(result) == len(region_ids)
    biases = [np.nan]
    for data in result.values():
        for stats in data.values():
            biases.append(stats["nmb"])
    mean_bias = np.nanmean(biases)
    assert mean_bias == pytest.approx(nmb_avg, abs=0.001, nan_ok=True)


@pytest.mark.parametrize(
    "freq,region_ids,data_freq,drop_stats",
    [
        ("yearly", {"EUROPE": "Europe"}, "monthly", ("mb", "mab")),
        ("yearly", {"EUROPE": "Europe"}, None, ("nmb",)),
    ],
)
@pytest.mark.filterwarnings("ignore:Mean of empty slice:RuntimeWarning")
def test__process_statistics_timeseries_drop_stats(
    example_coldata,
    freq: str,
    region_ids: dict[str, str],
    data_freq: str,
    drop_stats: tuple[str],
):
    result = _process_statistics_timeseries(
        data=example_coldata,
        freq=freq,
        region_ids=region_ids,
        use_weights=False,
        drop_stats=drop_stats,
        use_country=False,
        data_freq=data_freq,
    )
    assert result.keys().isdisjoint(drop_stats)


@pytest.mark.parametrize(
    "freq,region_ids,data_freq,exception,error",
    [
        pytest.param(
            "monthly",
            {"bla": "blub"},
            None,
            UnknownRegion,
            "no such region defined bla",
            id="unknown region",
        ),
        pytest.param(
            "daily",
            {},
            "daily",
            TemporalResolutionError,
            "failed to compute statistics timeseries, no co-located data available in specified base resolution daily",
            id="no daily data",
        ),
        pytest.param(
            "daily",
            {},
            "monthly",
            TemporalResolutionError,
            "Desired input frequency monthly is lower than desired output frequency daily",
            id="wrong data_freq",
        ),
    ],
)
def test__process_statistics_timeseries_error(
    example_coldata,
    freq: str,
    region_ids: dict[str, str],
    data_freq: str,
    exception: type[Exception],
    error: str,
):
    with pytest.raises(exception) as e:
        _process_statistics_timeseries(
            data=example_coldata,
            freq=freq,
            region_ids=region_ids,
            use_weights=False,
            drop_stats=(),
            use_country=False,
            data_freq=data_freq,
        )
    assert str(e.value) == error


@pytest.mark.parametrize(
    "freq,season,start,stop,min_yrs,station",
    [
        ("yearly", "all", 2000, 2015, 7, 1),
        ("yearly", "JJA", 2010, 2015, 4, 2),
        ("yearly", "SON", 2000, 2015, 7, 3),
        ("monthly", "all", 2000, 2015, 7, 2),
        ("monthly", "all", 2010, 2015, 4, 0),
    ],
)
@pytest.mark.parametrize("coldataset", ["fake_3d_trends"])
def test__make_trends(
    coldata: ColocatedData,
    freq: str,
    season: str,
    start: int,
    stop: int,
    min_yrs: int,
    station: int,
):
    obs_val = coldata.data.data[0, :, station]
    mod_val = coldata.data.data[1, :, station]
    time = coldata.data.time

    obs_trend, mod_trend = _make_trends(obs_val, mod_val, time, freq, season, start, stop, min_yrs)

    assert obs_trend["period"] == f"{start}-{stop}"
    assert mod_trend["period"] == f"{start}-{stop}"

    assert obs_trend["season"] == mod_trend["season"]

    trend_slops = [1, 2, 50, -3]

    eps = 0.1
    assert abs((obs_trend["m"] - trend_slops[station]) / obs_trend["m"]) < eps
    assert abs((mod_trend["m"] - trend_slops[station]) / mod_trend["m"]) < eps

    assert int(obs_trend["map_var"].split("_")[1]) == start
    assert int(mod_trend["map_var"].split("_")[1]) == start


@pytest.mark.parametrize(
    "freq,season,min_yrs,exception,error",
    [
        pytest.param(
            "daily",
            "JJA",
            4,
            ValueError,
            "daily",
            id="wrong freq",
        ),
        pytest.param(
            "yearly",
            "all",
            7,
            TrendsError,
            "min_yrs (7) larger than time between start and stop",
            id="wrong min_yrs",
        ),
    ],
)
@pytest.mark.parametrize("coldataset", ["fake_3d_trends"])
def test__make_trends_error(
    coldata: ColocatedData,
    freq: str,
    season: str,
    min_yrs: int,
    exception: type[Exception],
    error: str,
):
    obs_val = coldata.data.data[0, :, 0]
    mod_val = coldata.data.data[1, :, 0]
    time = coldata.data.time
    with pytest.raises(exception) as e:
        _make_trends(obs_val, mod_val, time, freq, season, 2010, 2015, min_yrs)
    assert str(e.value) == error


@pytest.mark.parametrize("coldataset", ["fake_3d_trends"])
def test__init_meta_glob(coldata: ColocatedData):
    # test the functionality of __init_meta_glob when there are KeyErrors. Other cases already covered
    no_meta_coldata = coldata.copy()
    no_meta_coldata.data.attrs = {}
    res = _init_meta_glob(no_meta_coldata)
    res.pop("processed_utc")
    assert set(res.values()) == {"UNDEFINED"}


@pytest.mark.parametrize("resolution", [("yearly")])
@pytest.mark.parametrize("coldataset", ["fake_3d_trends"])
def test__create_diurnal_weekly_data_object(coldata: ColocatedData, resolution: str):
    obj = _create_diurnal_weekly_data_object(coldata, resolution)
    assert isinstance(obj, xarray.Dataset)


@pytest.mark.parametrize(
    "resolution,error_message",
    [
        ("seasonal", "hour must not be empty"),
        ("cat", "Invalid resolution"),
    ],
)
@pytest.mark.parametrize("coldataset", ["fake_3d_trends"])
def test__create_diurnal_weekly_data_object__error(
    coldata: ColocatedData, resolution: str, error_message: str
):
    with pytest.raises(ValueError, match=error_message):
        _create_diurnal_weekly_data_object(coldata, resolution)


@pytest.mark.parametrize(
    "resolution",
    [
        ("seasonal"),
        ("yearly"),
    ],
)
def test__get_period_keys(resolution: str):
    res = _get_period_keys(resolution)
    assert np.all(isinstance(item, str) for item in res)


def test__map_indices():
    outer_idx = [2010, 2011, 2012, 2013, 2014, 2015]
    inner_idx = [2011, 2012]
    out = _map_indices(outer_idx, inner_idx)
    assert isinstance(out, np.ndarray)
    assert len(out) == len(outer_idx)


@pytest.mark.parametrize(
    "min_yrs, station_nb, colocate_time",
    [
        (0, 5, False),
        (0, 5, True),
        (4, 2, True),
        (4, 2, False),
        (3, 3, True),
        (3, 3, False),
    ],
)
def test__remove_less_covered(
    # coldata: dict,
    min_yrs: int,
    station_nb: int,
    colocate_time: bool,
):
    cd = (
        COLDATA["fake_3d_partial_trends_coltime"]()
        if colocate_time
        else COLDATA["fake_3d_partial_trends"]()
    )
    old_stations = cd.data.station_name.data

    new_cd = _remove_less_covered(cd, min_yrs)

    new_stations = new_cd.data.station_name.data

    if min_yrs == 0:
        assert len(old_stations) == len(new_stations)

    assert len(new_stations) == station_nb


def test__remove_less_covered_all():
    cd = COLDATA["fake_3d_partial_trends_coltime"]()
    new_cd = _remove_less_covered(cd, 1000)
    assert new_cd.data.station_name.shape[0] == 0
